# Copyright (C) 2022 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""A testsuite for the generators module."""

import inspect
import os
import unittest

import numpy as np
import pandas as pd
import tcpserver

from nemo import costs, generators

PORT = 9998
hydrogen_storage = generators.HydrogenStorage(1000, "H2 store")

dummy_arguments = {'axes': 0,
                   'build_limit': 1000,
                   'capacity': 100,
                   'capture': 0.85,
                   'column': 0,
                   'cost_per_mwh': 1000,
                   'discharge_hours': range(18, 21),
                   'efficiency': 30,
                   'filename': 'tracedata.csv',
                   'heatrate': 0.3,
                   'intensity': 0.7,
                   'kwh_per_litre': 10,
                   'label': 'a label',
                   'maxstorage': 1000,
                   'polygon': 31,
                   'rte': 0.8,
                   'self': None,
                   'shours': 8,
                   'solarmult': 2.5,
                   'tank': hydrogen_storage}

# This list should contain every generator class in generators.py.
# This ensures that the linters will not report any classes as unused
# because they do not feature in any scenario (eg, GreenPower). They
# are, however, tested below via Python introspection (and hence not
# named explicitly in source code).

classlist = [generators.Battery, generators.Behind_Meter_PV,
             generators.Biofuel, generators.Biomass,
             generators.Black_Coal, generators.CCGT,
             generators.CCGT_CCS, generators.CCS, generators.CST,
             generators.CSVTraceGenerator, generators.CentralReceiver,
             generators.Coal_CCS, generators.DemandResponse,
             generators.Diesel, generators.Electrolyser,
             generators.Fossil, generators.Fuelled,
             generators.Generator, generators.Geothermal,
             generators.Geothermal_EGS, generators.Geothermal_HSA,
             generators.GreenPower, generators.Hydro,
             generators.HydrogenGT, generators.HydrogenStorage,
             generators.OCGT, generators.PV, generators.PV1Axis,
             generators.ParabolicTrough, generators.Patch,
             generators.PumpedHydro, generators.Storage,
             generators.TraceGenerator, generators.Wind,
             generators.WindOffshore]


class TestGenerators(unittest.TestCase):
    """Test generators.py."""

    def setUp(self):
        """Test harness setup."""
        self.tracefile = 'tracedata.csv'
        with open(self.tracefile, 'w', encoding='utf-8') as tracefile:
            for i in range(100):
                print(f'{0.01 * i:.2f}, 0', file=tracefile)

        self.years = 1
        self.costs = costs.NullCosts()
        self.classes = [cls for cls in
                        inspect.getmembers(generators, inspect.isclass)
                        if cls[1].__module__ == generators.__name__]
        self.generators = []

        for (cls, clstype) in self.classes:
            # Skip abstract classes
            if cls in ['Generator', 'TraceGenerator',
                       'CSVTraceGenerator', 'Storage',
                       'HydrogenStorage']:
                continue

            # check that every class in generators.py is in classlist
            self.assertIn(clstype, classlist)

            args = inspect.getfullargspec(clstype.__init__).args
            arglist = []
            for arg in args:
                if dummy_arguments[arg] is not None:
                    arglist.append(dummy_arguments[arg])
            obj = clstype(*arglist)
            self.generators.append(obj)

    def tearDown(self):
        """Remove tracefile on teardown."""
        os.unlink(self.tracefile)

    def test_series(self):
        """Test series() method."""
        gen = generators.Generator(1, 0, 'label')
        # fake up these attributes
        gen.series_power = {1: 100}
        gen.series_spilled = {1: 200}
        # .. and then call gen.series()
        series1 = pd.Series(gen.series_power, dtype=float)
        self.assertTrue(gen.series()['power'].equals(other=series1))
        series2 = pd.Series(gen.series_spilled, dtype=float)
        self.assertTrue(gen.series()['spilled'].equals(other=series2))

    def test_step_abstract(self):
        """Test step() method in the abstract Generator class."""
        gen = generators.Generator(1, 0, 'label')
        with self.assertRaises(NotImplementedError):
            gen.step(0, 100)

    def test_step(self):
        """Test step() method."""
        for gen in self.generators:
            for hour in range(0, 10):
                gen.step(hour, 20)

    def test_store(self):
        """Test store() method."""
        for gen in self.generators:
            if gen.storage_p:
                for hour in range(0, 10):
                    gen.step(hour, 20)

    def test_capcost(self):
        """Test capcost() method."""
        for gen in self.generators:
            gen.capcost(self.costs)

    def test_capfactor_nan(self):
        """Test capfactor() NaN case."""
        gen = generators.Generator(1, 0, 'label')
        self.assertTrue(np.isnan(gen.capfactor()))

    def test_capfactor(self):
        """Test capfactor() method."""
        for gen in self.generators:
            # 10 MW for 10 hours = 100 MWh
            gen.series_power = {n: 10 for n in range(10)}
            self.assertEqual(gen.capfactor(), 10)

    def test_lcoe(self):
        """Test lcoe() method."""
        for gen in self.generators:
            # 10 MWh for 10 hours = 100 MWh
            gen.series_power = {n: 10 for n in range(10)}
            gen.lcoe(self.costs, self.years)

    def test_reset(self):
        """Test reset() method."""
        for gen in self.generators:
            gen.series_power = {n: 10 for n in range(10)}
            gen.series_spilled = {n: 10 for n in range(10)}
        for gen in self.generators:
            gen.reset()
        for gen in self.generators:
            self.assertEqual(len(gen.series_power), 0)
            self.assertEqual(len(gen.series_spilled), 0)

    def test_summary(self):
        """Test summary() method."""
        class MyContext:
            """A mocked up Context class."""

            costs = self.costs
            years = self.years

        context = MyContext()
        for gen in self.generators:
            gen.series_power = {n: 10 for n in range(10)}  # 10 MW * 10 h
            gen.series_spilled = {n: 1 for n in range(10)}  # 1 MW * 10 h
            # fake up a capcost() method for testing summary()
            gen.capcost = lambda costs: 100
            output = gen.summary(context)
            self.assertIn('capcost $100,', output)
            self.assertIn('supplied 100.00 MWh', output)
            self.assertIn('surplus 10.00 MWh', output)
            self.assertIn('CF 10.0%', output)

    def test_set_capacity(self):
        """Test set_capacity() method."""
        for gen in self.generators:
            gen.set_capacity(0.2)
        self.assertEqual(self.generators[0].capacity, 200)

    def test_set_storage(self):
        """Test set_storage() method."""
        # set_storage does not have a uniform calling convention for
        # different generator types, so handle each case specifically.
        for gen in self.generators:
            if isinstance(gen, generators.CST):
                testvalue = 10
                gen.set_storage(testvalue)
                self.assertEqual(gen.shours, testvalue)
                self.assertEqual(gen.maxstorage, gen.capacity * testvalue)
                self.assertEqual(gen.stored, 0.5 * gen.maxstorage)
            elif isinstance(gen, generators.Battery):
                testvalue = 4
                gen.set_storage(testvalue)
                self.assertEqual(gen.maxstorage, gen.capacity * testvalue)
                self.assertEqual(gen.stored, 0)

    def test_str(self):
        """Test __str__() method."""
        for gen in self.generators:
            str(gen)

    def test_repr(self):
        """Test __repr__() method."""
        for gen in self.generators:
            repr(gen)


class TestTraceGeneratorTimeout(unittest.TestCase):
    """Test timeout handling for a trace generator (Wind)."""

    def setUp(self):
        """Start the simple TCP server."""
        self.child = tcpserver.run(PORT, "block")
        self.url = f'http://localhost:{PORT}/data.csv'

    def tearDown(self):
        """Terminate TCP server on teardown."""
        self.child.terminate()

    def test_timeout(self):
        """Test fetching trace data from a dud server."""
        with self.assertRaises(TimeoutError):
            generators.Wind(1, 100, self.url, column=0)


class TestTraceGeneratorError(unittest.TestCase):
    """Test HTTP error handling for a trace generator."""

    def setUp(self):
        """Start the simple TCP server."""
        self.child = tcpserver.run(PORT, "http400")
        self.url = f'http://localhost:{PORT}/data.csv'

    def tearDown(self):
        """Terminate TCP server on teardown."""
        self.child.terminate()

    def test_timeout(self):
        """Test fetching trace data from a dud server."""
        with self.assertRaisesRegex(ConnectionError, "HTTP 400"):
            generators.Wind(1, 100, self.url, column=0)



# Copyright (C) 2022 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""A testsuite for the nem module."""

import configparser
import importlib
import unittest

import tcpserver

from nemo import nem

PORT = 9998


class MockConfigParser(configparser.ConfigParser):
    """A mocked up ConfigParser."""

    def httpget(self, section, option):
        """Fake the demand trace URL."""
        if (section, option) == ('demand', 'demand-trace'):
            return f'http://localhost:{PORT}/data.csv'
        return configparser.ConfigParser.get(self, section, option)

    def fileget(self, section, option):
        """Fake the demand trace filename."""
        if (section, option) == ('demand', 'demand-trace'):
            return 'nosuchfile.csv'
        return configparser.ConfigParser.get(self, section, option)


class TestDemandError(unittest.TestCase):
    """Test HTTP error handling for loading demand."""

    def setUp(self):
        """Start the simple TCP server."""
        self.child = tcpserver.run(PORT, "http400")
        self.oldget = configparser.ConfigParser.get
        configparser.ConfigParser.get = MockConfigParser.httpget

    def tearDown(self):
        """Terminate TCP server on teardown."""
        self.child.terminate()
        configparser.ConfigParser.get = self.oldget

    def test_http_error(self):
        """Test fetching demand data from a dud server."""
        with self.assertRaisesRegex(ConnectionError, "HTTP 400"):
            importlib.reload(nem)


class TestDemandTimeout(unittest.TestCase):
    """Test timeout handling when fetching demand data."""

    def setUp(self):
        """Start the simple TCP server."""
        self.child = tcpserver.run(PORT, "block")
        self.oldget = configparser.ConfigParser.get
        configparser.ConfigParser.get = MockConfigParser.httpget

    def tearDown(self):
        """Terminate TCP server on teardown."""
        self.child.terminate()
        configparser.ConfigParser.get = self.oldget

    def test_timeout(self):
        """Test fetching demand data from a dud server."""
        with self.assertRaises(TimeoutError):
            importlib.reload(nem)


class TestDemandNoSuchFile(unittest.TestCase):
    """Test timeout handling when opening non-existent file."""

    def setUp(self):
        """Set up the mock ConfigParser."""
        self.oldget = configparser.ConfigParser.get
        configparser.ConfigParser.get = MockConfigParser.fileget

    def tearDown(self):
        """Put the real ConfigParser back."""
        configparser.ConfigParser.get = self.oldget

    def test_timeout(self):
        """Test fetching demand data from a dud server."""
        with self.assertRaises(FileNotFoundError):
            importlib.reload(nem)



# Copyright (C) 2022 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""A testsuite for generators that have storage."""

import unittest

import pandas as pd

from nemo import configfile, generators
from nemo.polygons import WILDCARD


class TestStorage(unittest.TestCase):
    """Test Storage class."""

    def test_initialisation(self):
        """Test constructor."""
        storage = generators.Storage()
        self.assertEqual(storage.series_charge, {})

    def test_reset(self):
        """Test reset() method."""
        storage = generators.Storage()
        storage.series_charge = {0: 150}
        storage.reset()
        self.assertEqual(storage.series_charge, {})

    def test_record(self):
        """Test record() method."""
        storage = generators.Storage()
        storage.record(0, 100)
        storage.record(0, 50)
        storage.record(1, 75)
        self.assertEqual(storage.series_charge, {0: 150, 1: 75})

    def test_series(self):
        """Test series() method."""
        storage = generators.Storage()
        value = {0: 150}
        storage.series_charge = value
        series = pd.Series(value, dtype=float)
        self.assertTrue(storage.series()['charge'].equals(series))

    def test_store(self):
        """Test store() method."""
        storage = generators.Storage()
        with self.assertRaises(NotImplementedError):
            storage.store(0, 100)


class TestPumpedHydro(unittest.TestCase):
    """Test pumped hydro class in detail."""

    def setUp(self):
        """Test harness setup."""
        self.psh = generators.PumpedHydro(WILDCARD, 100, 1000, rte=1)

    def test_initialisation(self):
        """Test constructor."""
        self.assertTrue(self.psh.storage_p)
        self.assertEqual(self.psh.last_gen, None)
        self.assertEqual(self.psh.last_pump, None)
        self.assertEqual(self.psh.stored, 0.5 * self.psh.maxstorage)
        self.assertEqual(self.psh.rte, 1.0)
        self.assertEqual(self.psh.maxstorage, 1000)

    def test_series(self):
        """Test series() method."""
        series = self.psh.series()
        keys = series.keys()
        self.assertEqual(len(series), 3)
        self.assertTrue('power' in keys)
        self.assertTrue('spilled' in keys)
        self.assertTrue('charge' in keys)

    def test_pump_and_generate(self):
        """Test that pumping and generating cannot happen at the same time."""
        result = self.psh.store(hour=0, power=100)
        self.assertEqual(result, 100)
        result = self.psh.step(hour=0, demand=50)
        self.assertEqual(result, (0, 0))

    def test_pump_multiple(self):
        """Test that pump can run more than once per timestep."""
        result = self.psh.store(hour=0, power=100)
        self.assertEqual(result, 100)
        result = self.psh.store(hour=0, power=50)
        self.assertEqual(result, 0)
        self.assertEqual(self.psh.stored, 600)

    def test_step(self):
        """Test step() method."""
        for i in range(10):
            result = self.psh.step(hour=i, demand=50)
            self.assertEqual(result, (50, 0))
        self.assertEqual(self.psh.stored, 0)
        self.assertEqual(sum(self.psh.series_power.values()), 500)
        self.assertEqual(len(self.psh.series_power), 10)

    def test_store(self):
        """Test store() method."""
        result = self.psh.step(hour=0, demand=100)
        self.assertEqual(result, (100, 0))
        # Can't pump and generate in the same hour.
        result = self.psh.store(hour=0, power=250)
        self.assertEqual(result, 0)

        self.psh.stored = 800
        result = self.psh.store(hour=1, power=200)
        self.assertEqual(result, 100)
        self.assertEqual(self.psh.stored, 900)
        result = self.psh.store(hour=2, power=200)
        self.assertEqual(result, 100)
        self.assertEqual(self.psh.stored, 1000)
        result = self.psh.store(hour=3, power=200)
        self.assertEqual(result, 0)
        self.assertEqual(self.psh.stored, 1000)
        self.assertEqual(sum(self.psh.series_charge.values()), 200)
        self.assertEqual(len(self.psh.series_charge), 2)

    def test_store_multiple(self):
        """Test store() called multiple times."""
        # For now.

    def test_reset(self):
        """Test reset() method."""
        self.psh.series_power = {0: 200}
        self.psh.series_charge = {0: 150}
        self.psh.stored = 0
        self.psh.last_gen = 123
        self.psh.last_pump = 456
        self.psh.reset()
        self.assertEqual(self.psh.stored, 0.5 * self.psh.maxstorage)
        self.assertEqual(self.psh.last_gen, None)
        self.assertEqual(self.psh.last_pump, None)
        self.assertEqual(len(self.psh.series_charge), 0)
        self.assertEqual(len(self.psh.series_power), 0)


class TestCST(unittest.TestCase):
    """Test CST class in detail."""

    def setUp(self):
        """Test harness setup."""
        trace_file = configfile.get('generation', 'cst-trace')
        self.cst = generators.CentralReceiver(WILDCARD, 100, 2.5, 8,
                                              trace_file, 0)

    def test_initialisation(self):
        """Test CST constructor."""
        self.assertEqual(self.cst.capacity, 100)
        self.assertEqual(self.cst.shours, 8)
        self.assertEqual(self.cst.maxstorage,
                         self.cst.capacity * self.cst.shours)
        self.assertEqual(self.cst.stored, 0.5 * self.cst.maxstorage)
        self.assertEqual(self.cst.solarmult, 2.5)

    def test_set_capacity(self):
        """Test set_capacity() method."""
        self.cst.set_capacity(0.2)
        self.assertEqual(self.cst.capacity, 200)
        self.assertEqual(self.cst.maxstorage, 200 * self.cst.shours)

    def test_set_multiple(self):
        """Test set_multiple() method."""
        self.cst.set_multiple(3)
        self.assertEqual(self.cst.solarmult, 3)

    def test_set_storage(self):
        """Test set_storage() method."""
        self.cst.set_storage(6)
        self.assertEqual(self.cst.maxstorage,
                         self.cst.capacity * self.cst.shours)
        self.assertEqual(self.cst.stored, 0.5 * self.cst.maxstorage)

    def test_reset(self):
        """Test reset() method."""
        self.cst.stored = 0
        self.cst.reset()
        self.assertEqual(self.cst.stored, 0.5 * self.cst.maxstorage)


class TestBattery(unittest.TestCase):
    """Test Battery class in detail."""

    def test_initialisation(self):
        """Test Battery constructor."""
        batt = generators.Battery(WILDCARD, 400, 2, rte=1)
        self.assertEqual(batt.maxstorage, 800)
        self.assertEqual(batt.discharge_hours, range(18, 24))
        self.assertTrue(batt.empty_p())
        self.assertFalse(batt.full_p())
        self.assertTrue(batt.storage_p)
        self.assertEqual(batt.rte, 1)
        self.assertEqual(batt.stored, 0)
        self.assertEqual(batt.runhours, 0)
        self.assertEqual(len(batt.series_charge), 0)

    def test_series(self):
        """Test series() method."""
        batt = generators.Battery(WILDCARD, 400, 2)
        series = batt.series()
        keys = series.keys()
        self.assertEqual(len(series), 3)
        self.assertTrue('power' in keys)
        self.assertTrue('spilled' in keys)
        self.assertTrue('charge' in keys)

    def test_empty_p(self):
        """Test the empty_p() method."""
        batt = generators.Battery(WILDCARD, 800, 1, rte=1)
        self.assertEqual(batt.stored, 0)
        self.assertTrue(batt.empty_p())

    def test_full_p(self):
        """Test the full_p() method."""
        batt = generators.Battery(WILDCARD, 800, 1, rte=1)
        self.assertFalse(batt.full_p())
        batt.stored = 800
        self.assertTrue(batt.full_p())

    def test_discharge(self):
        """Test discharging outside of discharge hours."""
        # Test discontiguous hour range
        hrs = [0, 1] + list(range(18, 24))
        batt = generators.Battery(WILDCARD, 400, 8, discharge_hours=hrs, rte=1)
        batt.stored = 400
        for hour in range(24):
            result = batt.step(hour=hour, demand=50)
            # 0,0 if no discharging permitted, 50,0 otherwise
            self.assertEqual(result, (50, 0) if hour in hrs else (0, 0),
                             f'discharge failed in hour {hour}')
        self.assertTrue(batt.empty_p())

    def test_charge(self):
        """Test (normal) charging inside and outside of discharge hours."""
        # Test discontiguous hour range
        hrs = [0, 1] + list(range(18, 24))
        batt = generators.Battery(WILDCARD, 400, 8, discharge_hours=hrs, rte=1)
        for hour in range(24):
            result = batt.store(hour=hour, power=50)
            # 0 if no charging permitted, 50 otherwise
            self.assertEqual(result, 0 if hour in hrs else 50)
        nhours = 24 - len(hrs)
        self.assertEqual(batt.stored, 50 * nhours)
        self.assertEqual(sum(batt.series_charge.values()), 50 * nhours)

    def test_charge_multiple(self):
        """Test multiple calls to store()."""
        # 125 MW x 8h = 1000 MWh
        batt = generators.Battery(WILDCARD, 125, 8, discharge_hours=[], rte=1)
        result = batt.store(hour=12, power=100)
        self.assertEqual(result, 100)
        result = batt.store(hour=12, power=100)
        self.assertEqual(result, 25)
        result = batt.store(hour=12, power=100)
        self.assertEqual(result, 0)
        self.assertEqual(batt.stored, 125)

    def test_to_full(self):
        """Test charging to full."""
        batt = generators.Battery(WILDCARD, 400, 2, rte=1)
        batt.stored = 700
        result = batt.store(hour=1, power=200)
        self.assertEqual(result, 100)
        self.assertEqual(batt.stored, 800)

    def test_zero_power(self):
        """Test a battery with zero capacity."""
        batt = generators.Battery(WILDCARD, 0, 1)
        result = batt.store(hour=0, power=400)
        self.assertEqual(result, 0)
        result = batt.step(hour=0, demand=400)
        self.assertEqual(result, (0, 0))
        self.assertEqual(len(batt.series_charge), 0)
        self.assertEqual(batt.runhours, 0)

    def test_reset(self):
        """Test battery reset() method."""
        batt = generators.Battery(WILDCARD, 400, 2, rte=1)
        batt.series_power = {0: 200}
        batt.series_charge = {0: 150}
        batt.reset()
        self.assertEqual(len(batt.series_charge), 0)
        self.assertEqual(len(batt.series_power), 0)

    def test_round_trip_efficiency(self):
        """Test a battery with 50% round trip efficiency."""
        batt = generators.Battery(WILDCARD, 100, 4, rte=0.5)
        self.assertEqual(batt.stored, 0)
        result = batt.store(hour=0, power=100)
        self.assertEqual(result, 100)
        self.assertEqual(batt.stored, 100)
        result = batt.step(hour=18, demand=1000)
        self.assertEqual(result, (50, 0))
        self.assertEqual(batt.stored, 50)


class TestElectrolyser(unittest.TestCase):
    """Test Electrolyser class in detail."""

    def setUp(self):
        """Test harness setup."""
        self.tank = generators.HydrogenStorage(400, 'test')
        self.electrolyser = \
            generators.Electrolyser(self.tank, WILDCARD, 100,
                                    efficiency=1)

    def test_type_error(self):
        """Check that the wrong type raises a TypeError."""
        with self.assertRaises(TypeError):
            generators.Electrolyser(None, 1, 100, 'test')

    def test_series(self):
        """Test series() method."""
        series = self.electrolyser.series()
        keys = series.keys()
        self.assertEqual(len(series), 3)
        self.assertTrue('power' in keys)
        self.assertTrue('spilled' in keys)
        self.assertTrue('charge' in keys)

    def test_step(self):
        """Test step() method."""
        # an electrolyser is not a generator
        self.assertEqual(self.electrolyser.step(0, 100), (0, 0))
        # store 100 MWh of hydrogen
        self.assertEqual(self.electrolyser.store(0, 100), 100)
        # store another 100 MWh of hydrogen
        self.assertEqual(self.electrolyser.store(0, 100), 100)
        # tank is full, none stored
        self.assertEqual(self.electrolyser.store(0, 100), 0)



# Copyright (C) 2017, 2019 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""The core of the simulation engine."""

from math import isclose

import numpy as np
import pandas as pd

from nemo import regions


def _sim(context, date_range):
    # reset generator internal state
    for gen in context.generators:
        gen.reset()

    generation = np.zeros((len(date_range), len(context.generators)))
    spill = np.zeros((len(date_range), len(context.generators)))

    # Extract generators in the regions of interest.
    gens = [g for g in context.generators if g.region() in context.regions]

    # Zero out polygon demands we don't care about.
    for rgn in [r for r in regions.All if r not in context.regions]:
        for poly in rgn.polygons:
            context.demand[poly - 1] = 0

    # We are free to scribble all over demand_copy. Use ndarray for speed.
    demand_copy = context.demand.copy().values
    residual_demand = demand_copy.sum(axis=1)

    for hour, date in enumerate(date_range):
        hour_demand = demand_copy[hour]
        residual_hour_demand = residual_demand[hour]

        if context.verbose:
            print('STEP:', date)
            print('DEMAND:', {a: round(b, 2) for a, b in
                              enumerate(hour_demand)})

        _dispatch(context, hour, residual_hour_demand, gens, generation, spill)

        if context.verbose:
            print('ENDSTEP:', date)

    # Change the numpy arrays to dataframes for human consumption
    context.generation = pd.DataFrame(index=date_range, data=generation)
    context.spill = pd.DataFrame(index=date_range, data=spill)


def _store_spills(context, hour, gen, generators, spl):
    """Store spills from a generator into any storage."""
    assert spl > 0, f'{spl} is <= 0'
    for other in list(g for g in generators if g.storage_p):
        stored = other.store(hour, spl)
        spl -= stored
        if spl < 0 and isclose(spl, 0, abs_tol=1e-6):
            spl = 0
        assert spl >= 0

        # energy stored <= energy transferred, according to store's RTE
        if context.verbose:
            # show the energy transferred, not stored
            print('STORE:', gen, '->', other, f'({stored:.1f})')

        if spl == 0:
            # early exit
            break
    return spl


def _dispatch(context, hour, residual_hour_demand, gens, generation, spill):
    """Dispatch power from each generator in merit (list) order."""
    # async_demand is the maximum amount of the demand in this
    # hour that can be met from non-synchronous
    # generation. Non-synchronous generation in excess of this
    # value must be spilled.
    async_demand = residual_hour_demand * context.nsp_limit

    for gidx, generator in enumerate(gens):
        if not generator.synchronous_p and async_demand < residual_hour_demand:
            gen, spl = generator.step(hour, async_demand)
        else:
            gen, spl = generator.step(hour, residual_hour_demand)
        assert gen < residual_hour_demand or \
            isclose(gen, residual_hour_demand), \
            f"generation ({gen:.4f}) > demand " + \
            f"({residual_hour_demand:.4f}) for {generator}"
        generation[hour, gidx] = gen

        if not generator.synchronous_p:
            async_demand -= gen
            assert async_demand > 0 or isclose(async_demand, 0, abs_tol=1e-6)
            async_demand = max(0, async_demand)

        residual_hour_demand -= gen
        assert residual_hour_demand > 0 or \
            isclose(residual_hour_demand, 0, abs_tol=1e-6)
        residual_hour_demand = max(0, residual_hour_demand)

        if context.verbose:
            print(f'GENERATOR: {generator},',
                  f'generation: {gen:.1f}',
                  f'spill: {spl:.1f}',
                  f'residual-demand: {residual_hour_demand:.1f}',
                  f'async-demand: {async_demand:.1f}')

        if spl > 0:
            spill[hour, gidx] = \
                _store_spills(context, hour, generator, gens, spl)


def run(context, starthour=None, endhour=None):
    """Run the simulation."""
    if not isinstance(context.regions, list):
        raise TypeError

    if starthour is None:
        starthour = context.demand.index.min()
    if endhour is None:
        endhour = context.demand.index.max()
    date_range = pd.date_range(starthour, endhour, freq='H')

    _sim(context, date_range)

    # Calculate unserved energy.
    agg_demand = context.demand.sum(axis=1)
    agg_generation = context.generation.sum(axis=1)
    unserved = agg_demand - agg_generation
    # Ignore unserved events very close to 0 (rounding errors)
    context.unserved = unserved[~np.isclose(unserved, 0)]



# Copyright (C) 2021 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""Demand side scenarios."""

import numpy as np
import pandas as pd


def roll(label):
    """roll:X rolls the load by X timesteps.

    >>> roll("roll:3")  # doctest: +ELLIPSIS
    <function roll.<locals>.<lambda> at ...>
    >>> roll("junk string")
    Traceback (most recent call last):
    AssertionError
    """
    assert label.startswith('roll:')

    # label form: "roll:X" rolls the load by X timesteps
    _, posns = label.split(':')
    posns = int(posns)
    return lambda context: _roll_demand(context, posns)


def scale(label):
    """scale:X scales all of the load uniformly by X%.

    >>> scale("scale:10")  # doctest: +ELLIPSIS
    <function scale.<locals>.<lambda> at ...>
    >>> scale("junkstring")
    Traceback (most recent call last):
    AssertionError
    """
    assert label.startswith('scale:')

    _, factor = label.split(':')
    factor = 1 + float(factor) / 100
    return lambda context: _scale_demand_by(context, factor)


def scalex(label):
    """scalex:H1:H2:X scales hours H1 to H2 by X%.

    >>> scalex("scalex:8:12:10")  # doctest: +ELLIPSIS
    <function scalex.<locals>.<lambda> at ...>
    >>> scalex("scalex:10:30:5")
    Traceback (most recent call last):
    ValueError: hour > 24
    >>> scalex("scalex:12:8:5")
    Traceback (most recent call last):
    ValueError: to_hour comes before from_hour
    >>> scalex("junkstring")
    Traceback (most recent call last):
    AssertionError
    """
    assert label.startswith('scalex:')

    _, hour1, hour2, factor = label.split(':')
    from_hour = int(hour1)
    to_hour = int(hour2)
    if from_hour < 0 or to_hour < 0:
        raise ValueError("hour < 0")
    if from_hour > 24 or to_hour > 24:
        raise ValueError("hour > 24")
    if to_hour <= from_hour:
        raise ValueError("to_hour comes before from_hour")
    factor = 1 + float(factor) / 100
    return lambda context: _scale_range_demand(context,
                                               from_hour, to_hour, factor)


def scaletwh(label):
    """scaletwh:N scales demand to N TWh.

    >>> scaletwh("scaletwh:100")  # doctest: +ELLIPSIS
    <function scaletwh.<locals>.<lambda> at ...>
    >>> scaletwh("junkstring")
    Traceback (most recent call last):
    AssertionError
    """
    assert label.startswith('scaletwh:')

    _, val = label.split(':')
    new_demand = float(val)
    return lambda context: _scale_demand_twh(context, new_demand)


def shift(label):
    """shift:N:H1:H2 shifts N MW of daily load from H1 to H2.

    >>> shift("shift:3:10:14")  # doctest: +ELLIPSIS
    <function shift.<locals>.<lambda> at ...>
    >>> shift("junkstring")
    Traceback (most recent call last):
    AssertionError
    """
    assert label.startswith('shift:')
    _, demand, hour1, hour2 = label.split(':')
    demand = int(demand)
    from_hour = int(hour1)
    to_hour = int(hour2)
    if from_hour < 0 or to_hour < 0:
        raise ValueError("hour < 0")
    if from_hour > 24 or to_hour > 24:
        raise ValueError("hour > 24")
    return lambda context: _shift_demand(context, demand, from_hour, to_hour)


def peaks(label):
    """peaks:N:X reduces demand peaks over N MW by X%.

    >>> peaks("peaks:5:10")  # doctest: +ELLIPSIS
    <function peaks.<locals>.<lambda> at ...>
    >>> peaks("junkstring")
    Traceback (most recent call last):
    AssertionError
    """
    assert label.startswith('peaks:')
    _, power, factor = label.split(':')
    power = int(power)
    factor = 1 + float(factor) / 100
    return lambda context: _scale_peaks(context, power, factor)


def npeaks(label):
    """npeaks:N:X adjusts top N demand peaks by X%.

    >>> npeaks("npeaks:5:10")  # doctest: +ELLIPSIS
    <function npeaks.<locals>.<lambda> at ...>
    >>> npeaks("junkstring")
    Traceback (most recent call last):
    AssertionError
    """
    assert label.startswith('npeaks:')
    _, topn, factor = label.split(':')
    topn = int(topn)
    factor = 1 + float(factor) / 100
    return lambda context: _scale_npeaks(context, topn, factor)


def unchanged(_):
    """No demand modification.

    >>> unchanged("any")  # doctest: +ELLIPSIS
    <function unchanged.<locals>.<lambda> at ...>
    """
    return lambda context: context


switch_table = [
    (lambda s: s == 'unchanged', unchanged),
    (lambda s: s.startswith('roll:'), roll),
    (lambda s: s.startswith('scale:'), scale),
    (lambda s: s.startswith('scalex:'), scalex),
    (lambda s: s.startswith('scaletwh:'), scaletwh),
    (lambda s: s.startswith('shift:'), shift),
    (lambda s: s.startswith('peaks:'), peaks),
    (lambda s: s.startswith('npeaks:'), npeaks),
]


# Demand modifiers
def switch(label):
    """Return a callback function to modify the demand.

    >>> switch('unchanged')	  # doctest: +ELLIPSIS
    <function ...>

    >>> switch('roll:10')  # doctest: +ELLIPSIS
    <function ...>

    >>> switch('scale:5')  # doctest: +ELLIPSIS
    <function ...>

    >>> switch('scalex:0:10:5')  # doctest: +ELLIPSIS
    <function ...>

    >>> switch('shift:100:10:12')  # doctest: +ELLIPSIS
    <function ...>

    >>> switch('shift:100:-2:12')  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ValueError: hour < 0

    >>> switch('shift:100:12:25')  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ValueError: hour > 24

    >>> switch('scalex:-1:12:20')  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ValueError: hour < 0

    >>> switch('scalex:12:25:20')  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ValueError: hour > 24

    >>> switch('scalex:20:8:20')  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ValueError: to_hour comes before from_hour

    >>> switch('peaks:10:34000')  # doctest: +ELLIPSIS
    <function ...>

    >>> switch('npeaks:10:5')  # doctest: +ELLIPSIS
    <function ...>

    >>> switch('foo')
    Traceback (most recent call last):
    ValueError: invalid scenario: foo
    """
    for (predicate, callback) in switch_table:
        if predicate(label):
            return callback(label)

    raise ValueError(f'invalid scenario: {label}')


def _roll_demand(context, posns):
    """
    Roll demand by posns timesteps.

    >>> c = type('context', (), {})
    >>> c.demand = pd.DataFrame(list(range(10)))
    >>> _roll_demand(c, 1)
    >>> print(c.demand)
       0
    0  9
    1  0
    2  1
    3  2
    4  3
    5  4
    6  5
    7  6
    8  7
    9  8
    """
    idx = context.demand.index
    values = np.roll(context.demand.values, posns)
    context.demand = pd.DataFrame(data=values, index=idx)


def _scale_range_demand(context, from_hour, to_hour, factor):
    """
    Scale demand between from_hour and to_hour by factor%.

    >>> c = type('context', (), {})
    >>> c.demand = pd.DataFrame(list(range(10)))
    >>> _scale_range_demand(c, 0, 4, 1.2)
    >>> print(c.demand)
         0
    0  0.0
    1  1.2
    2  2.4
    3  3.6
    4  4.0
    5  5.0
    6  6.0
    7  7.0
    8  8.0
    9  9.0
    """
    for hour in range(from_hour, to_hour):
        context.demand[hour::24] *= factor


def _scale_demand_twh(context, new_demand):
    """
    Scale demand to new_demand TWh.

    >>> c = type('context', (), {})
    >>> c.demand = pd.DataFrame([100]*10)
    >>> _scale_demand_twh(c, 0.0002)
    >>> print(c.demand.loc[0])
    0    20.0
    Name: 0, dtype: float64
    """
    total_demand = context.demand.values.sum()
    new_demand *= 10 ** 6
    context.demand *= new_demand / total_demand


def _scale_demand_by(context, factor):
    """
    Scale demand by factor%.

    >>> c = type('context', (), {})
    >>> c.demand = pd.DataFrame([0, 1, 2])
    >>> _scale_demand_by(c, 1.2)
    >>> print(c.demand)
         0
    0  0.0
    1  1.2
    2  2.4
    """
    context.demand *= factor


def _shift_demand(context, demand, from_hour, to_hour):
    """Move n MW of demand from from_hour to to_hour.

    >>> from nemo import context, regions
    >>> ctx = context.Context()
    >>> ctx.regions = [regions.sa]
    >>> ctx.demand = np.zeros((43, 24 * 7))
    >>> ctx.demand[26][0::24] = 100  # polygon 27
    >>> ctx.demand[31][0::24] = 100  # polygon 32
    >>> saved_sum = ctx.demand.sum()
    >>> _shift_demand(ctx, 50, 0, 12) # shift 50MW from midnight to noon
    >>> assert ctx.demand.sum() == saved_sum  # verify no change
    >>> ctx.demand[26][0], ctx.demand[26][12]  # 5MW -> noon
    (95.0, 5.0)
    >>> ctx.demand[31][0], ctx.demand[31][12]  # 45MW -> noon
    (55.0, 45.0)
    """
    # Shift demand within in each polygon
    for poly in range(43):
        for regn in context.regions:
            if poly + 1 in regn.polygons:
                weight = regn.polygons[poly + 1]
                if context.demand[poly].sum() > 0:
                    context.demand[poly, from_hour::24] -= demand * weight
                    context.demand[poly, to_hour::24] += demand * weight
    assert np.all(context.demand >= 0), \
        f"negative load in hour {from_hour}"


def _scale_peaks(context, power, factor):
    """
    Adjust demand peaks over N megawatts by factor%.

    >>> c = type('context', (), {})
    >>> c.demand = pd.DataFrame([[0.0]*5]*5)
    >>> c.demand.loc[3] = 5000
    >>> _scale_peaks(c, 3000, 0.5)
    >>> c.demand.loc[3]
    0    2500.0
    1    2500.0
    2    2500.0
    3    2500.0
    4    2500.0
    Name: 3, dtype: float64
    """
    agg_demand = context.demand.sum(axis=1)
    context.demand[agg_demand > power] *= factor


def _scale_npeaks(context, topn, factor):
    """
    Adjust top N demand peaks by X%.

    >>> c = type('context', (), {})
    >>> c.demand = pd.DataFrame([[0.0]*5]*5)
    >>> c.demand.loc[3] = 5000
    >>> c.demand.loc[4] = 3000
    >>> _scale_npeaks(c, 1, 0.5)
    >>> c.demand.loc[4]
    0    3000.0
    1    3000.0
    2    3000.0
    3    3000.0
    4    3000.0
    Name: 4, dtype: float64
    >>> c.demand.loc[3]
    0    2500.0
    1    2500.0
    2    2500.0
    3    2500.0
    4    2500.0
    Name: 3, dtype: float64
    """
    agg_demand = context.demand.sum(axis=1).sort_values(ascending=False)
    rng = agg_demand.head(topn).index
    context.demand.loc[rng] *= factor



# Copyright (C) 2022 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""A testsuite for the utils module."""

# Some protected members (eg _figure) are accessed to facilitate testing.
# pylint: disable=protected-access

import os
import unittest

import pytest

from nemo import context, scenarios, sim, utils


class TestUtils(unittest.TestCase):
    """Tests for utils.py functions."""

    def setUp(self):
        """Test harness setup."""
        self.context = context.Context()
        scenarios.ccgt(self.context)
        sim.run(self.context)

    @pytest.mark.mpl_image_compare
    def test_figure_1(self):
        """Test simple supply/demand plot."""
        utils._figure(self.context, spills=True, showlegend=True, xlim=None)
        return utils.plt.gcf()

    @pytest.mark.mpl_image_compare
    def test_figure_2(self):
        """Test supply/demand plot with many generators."""
        self.context.generators *= 2  # double list length
        sim.run(self.context)
        utils._figure(self.context, spills=True, showlegend=True, xlim=None)
        return utils.plt.gcf()

    def test_plot_1(self):
        """Test plot() function writing to file."""
        fname = 'test_plot_1.png'
        try:
            os.unlink(fname)
        except FileNotFoundError:
            pass
        utils.plot(self.context, filename=fname)
        try:
            os.stat(fname)
            exists = True
        except FileNotFoundError:
            exists = False
        self.assertTrue(exists)
        os.unlink(fname)



# Copyright (C) 2017 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""The National Electricity Market Optimiser (NEMO)."""

import nemo.nem  # noqa: F401
from nemo.context import Context
from nemo.sim import run
from nemo.utils import plot

__all__ = ['Context', 'run', 'plot']



# Copyright (C) 2012, 2013, 2014, 2022 Ben Elliston
# Copyright (C) 2014, 2015, 2016 The University of New South Wales
# Copyright (C) 2016 IT Power (Australia)
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""Supply side scenarios."""

from nemo import configfile, regions
from nemo.generators import (CCGT, CCGT_CCS, CST, OCGT, Battery, Biofuel,
                             Black_Coal, CentralReceiver, Coal_CCS,
                             DemandResponse, Hydro, PumpedHydro, PV1Axis, Wind,
                             WindOffshore)
from nemo.polygons import (WILDCARD, cst_limit, offshore_wind_limit, pv_limit,
                           wind_limit)
from nemo.types import UnreachableError


def _demand_response():
    """Return a list of DR 'generators'."""
    dr1 = DemandResponse(WILDCARD, 1000, 100, "DR100")
    dr2 = DemandResponse(WILDCARD, 1000, 500, "DR500")
    dr3 = DemandResponse(WILDCARD, 1000, 1000, "DR1000")
    return [dr1, dr2, dr3]


def _hydro():
    """Return a list of existing hydroelectric generators."""
    hydro24 = Hydro(24, 42.5, label='poly 24 hydro')
    hydro31 = Hydro(31, 43, label='poly 31 hydro')
    hydro35 = Hydro(35, 71, label='poly 35 hydro')
    hydro36 = Hydro(36, 2513.9, label='poly 36 hydro')
    hydro38 = Hydro(38, 450, label='poly 38 hydro')
    hydro39 = Hydro(39, 13.8, label='poly 39 hydro')
    hydro40 = Hydro(40, 586.6, label='poly 40 hydro')
    hydro41 = Hydro(41, 280, label='poly 41 hydro')
    hydro42 = Hydro(42, 590.4, label='poly 42 hydro')
    hydro43 = Hydro(43, 462.5, label='poly 43 hydro')

    # Pumped hydro
    # QLD: Wivenhoe (http://www.csenergy.com.au/content-%28168%29-wivenhoe.htm)
    psh17 = PumpedHydro(17, 500, 5000, label='poly 17 pumped-hydro')
    # NSW: Tumut 3 (6x250), Bendeela (2x80) and Kangaroo Valley (2x40)
    psh36 = PumpedHydro(36, 1740, 15000, label='poly 36 pumped-hydro')
    return [psh17, psh36] + \
        [hydro24, hydro31, hydro35, hydro36, hydro38, hydro39] + \
        [hydro40, hydro41, hydro42, hydro43]


def replacement(context):
    """Replace the current NEM fleet, more or less."""
    context.generators = \
        [Black_Coal(WILDCARD, 0)] + _hydro() + [OCGT(WILDCARD, 0)]


def _one_ccgt(context):
    """One CCGT only."""
    context.generators = [CCGT(WILDCARD, 0)]


def ccgt(context):
    """All gas scenario."""
    context.generators = [CCGT(WILDCARD, 0)] + _hydro() + [OCGT(WILDCARD, 0)]


def ccgt_ccs(context):
    """CCGT CCS scenario."""
    # pylint: disable=redefined-outer-name
    ccgt = CCGT_CCS(WILDCARD, 0)
    ocgt = OCGT(WILDCARD, 0)
    context.generators = [ccgt] + _hydro() + [ocgt]


def coal_ccs(context):
    """Coal CCS scenario."""
    coal = Coal_CCS(WILDCARD, 0)
    ocgt = OCGT(WILDCARD, 0)
    context.generators = [coal] + _hydro() + [ocgt]


def _every_poly(gentype):
    """Create a generator of type gentype in each of the 44 polygons."""
    result = []
    for poly in range(1, 44):
        if gentype == Biofuel:
            result.append(gentype(poly, 0, label=f'polygon {poly} GT'))
        elif gentype == PV1Axis:
            cfg = configfile.get('generation', 'pv1axis-trace')
            result.append(gentype(poly, 0, cfg, poly - 1,
                                  build_limit=pv_limit[poly],
                                  label=f'polygon {poly} PV'))
        elif gentype == CentralReceiver:
            cfg = configfile.get('generation', 'cst-trace')
            result.append(gentype(poly, 0, 2.0, 6, cfg, poly - 1,
                                  build_limit=cst_limit[poly],
                                  label=f'polygon {poly} CST'))
        elif gentype == Wind:
            cfg = configfile.get('generation', 'wind-trace')
            result.append(gentype(poly, 0, cfg, poly - 1,
                                  build_limit=wind_limit[poly],
                                  label=f'polygon {poly} wind'))
    return result


def re100(context):
    """100% renewable electricity."""
    result = []
    # The following list is in merit order.
    for g in [PV1Axis, Wind, WindOffshore, PumpedHydro, Hydro,
              CentralReceiver, Biofuel]:
        if g == PumpedHydro:
            result += [h for h in _hydro() if isinstance(h, PumpedHydro)]
        elif g == Hydro:
            result += [h for h in _hydro() if not isinstance(h, PumpedHydro)]
        elif g == WindOffshore:
            cfg = configfile.get('generation', 'offshore-wind-trace')
            for column, poly in enumerate([31, 36, 38, 40]):
                result.append(g(poly, 0, cfg, column,
                                build_limit=offshore_wind_limit[poly],
                                label='polygon {poly} offshore'))
        elif g in [Biofuel, PV1Axis, CentralReceiver, Wind]:
            result += _every_poly(g)
        else:
            raise UnreachableError('unhandled generator type')
    context.generators = result


def re100_batteries(context):
    """Use lots of renewables plus battery storage."""
    re100(context)
    # discharge between 6pm and 6am daily
    hrs = list(range(0, 7)) + list(range(18, 24))
    battery = Battery(WILDCARD, 0, 4, discharge_hours=hrs)
    context.generators.insert(0, battery)


def _one_per_poly(region):
    """Return three lists of wind, PV and CST generators, one per polygon."""
    pv = []
    wind = []
    cst = []

    wind_cfg = configfile.get('generation', 'wind-trace')
    pv_cfg = configfile.get('generation', 'pv1axis-trace')
    cst_cfg = configfile.get('generation', 'cst-trace')

    for poly in region.polygons:
        wind.append(Wind(poly, 0, wind_cfg,
                         poly - 1,
                         build_limit=wind_limit[poly],
                         label=f'poly {poly} wind'))
        pv.append(PV1Axis(poly, 0, pv_cfg,
                          poly - 1,
                          build_limit=pv_limit[poly],
                          label=f'poly {poly} PV'))
        cst.append(CentralReceiver(poly, 0, 2.5, 8, cst_cfg,
                                   poly - 1,
                                   build_limit=cst_limit[poly],
                                   label=f'poly {poly} CST'))
    return wind, pv, cst


def re100_one_region(context, region):
    """100% renewables in one region only."""
    re100(context)
    context.regions = [region]
    wind, pv, cst = _one_per_poly(region)
    newlist = wind
    newlist += pv
    newlist += [g for g in context.generators if
                isinstance(g, Hydro) and g.region() is region]
    newlist += cst
    newlist += [g for g in context.generators if
                isinstance(g, Biofuel) and g.region() is region]
    context.generators = newlist


def re_plus_ccs(context):
    """Mostly renewables with fossil and CCS augmentation."""
    re100(context)
    coal = Black_Coal(WILDCARD, 0)
    # pylint: disable=redefined-outer-name
    coal_ccs = Coal_CCS(WILDCARD, 0)
    # pylint: disable=redefined-outer-name
    ccgt = CCGT(WILDCARD, 0)
    ccgt_ccs = CCGT_CCS(WILDCARD, 0)
    ocgt = OCGT(WILDCARD, 0)
    context.generators = [coal, coal_ccs, ccgt, ccgt_ccs] + \
        context.generators[:-4] + [ocgt]


def re_plus_fossil(context):
    """Mostly renewables with some fossil augmentation."""
    re100(context)
    context.generators = \
        [Black_Coal(WILDCARD, 0), CCGT(WILDCARD, 0)] + \
        context.generators[:-4] + [OCGT(WILDCARD, 0)]


def re100_dsp(context):
    """Mostly renewables with demand side participation."""
    re100(context)
    context.generators += _demand_response()


def re100_nocst(context):
    """100% renewables, but no CST."""
    re100(context)
    newlist = [g for g in context.generators if not isinstance(g, CST)]
    context.generators = newlist


def re100_nsw(context):
    """100% renewables in New South Wales only."""
    re100_one_region(context, regions.nsw)


def re100_qld(context):
    """100% renewables in Queensland only."""
    re100_one_region(context, regions.qld)


def re100_south_aus(context):
    """100% renewables in South Australia only."""
    re100_one_region(context, regions.sa)


supply_scenarios = {'__one_ccgt__': _one_ccgt,  # nb. for testing only
                    'ccgt': ccgt,
                    'ccgt-ccs': ccgt_ccs,
                    'coal-ccs': coal_ccs,
                    're+ccs': re_plus_ccs,
                    're+fossil': re_plus_fossil,
                    're100': re100,
                    're100-qld': re100_qld,
                    're100-nsw': re100_nsw,
                    're100-sa': re100_south_aus,
                    're100+batteries': re100_batteries,
                    're100+dsp': re100_dsp,
                    're100-nocst': re100_nocst,
                    'replacement': replacement}



# Copyright (C) 2022 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""A testsuite for the configfile module."""

import configparser
import importlib
import os
import unittest

from nemo import configfile


class TestConfigfile(unittest.TestCase):
    """Tests for configfile.py functions."""

    def test_get(self):
        """Test get() function."""
        self.assertEqual(configfile.get('optimiser', 'sigma'), '2.0')
        with self.assertRaises(configparser.NoSectionError):
            configfile.get('nosection', 'sigma')
        with self.assertRaises(configparser.NoOptionError):
            configfile.get('optimiser', 'nooption')

    def test_has_option_p(self):
        """Test has_option_p() function."""
        self.assertTrue(configfile.has_option_p('optimiser', 'sigma'))
        self.assertFalse(configfile.has_option_p('optimiser', 'nooption'))
        self.assertFalse(configfile.has_option_p('nosection', 'sigma'))


class TestConfigFileSet(unittest.TestCase):
    """Test with NEMORC set."""

    def setUp(self):
        """Set NEMORC."""
        try:
            self.old = os.environ['NEMORC']
        except KeyError:
            self.old = None
        os.environ['NEMORC'] = '/file/not/found'

    def tearDown(self):
        """Clean up the environment."""
        if self.old is None:
            del os.environ['NEMORC']
        else:
            os.environ['NEMORC'] = self.old
        importlib.reload(configfile)

    def test_open(self):
        """Test opening a non-existent config file."""
        with self.assertRaises(FileNotFoundError):
            importlib.reload(configfile)


class TestConfigFileUnset(unittest.TestCase):
    """Test with NEMORC unset."""

    def setUp(self):
        """Unset NEMORC."""
        try:
            self.old = os.environ['NEMORC']
            del os.environ['NEMORC']
        except KeyError:
            self.old = None
        assert 'NEMORC' not in os.environ
        importlib.reload(configfile)

    def tearDown(self):
        """Reset the environment."""
        if self.old:
            os.environ['NEMORC'] = self.old
        importlib.reload(configfile)

    def test_open(self):
        """Test opening the default configuration file."""
        reference = configparser.ConfigParser()
        reference.read('nemo.cfg')
        self.assertEqual(reference, configfile.config)



# Copyright (C) 2022 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""A testsuite for the Context class."""

import unittest

import numpy as np
import pandas as pd

import nemo
from nemo import regions


class TestContextMethods(unittest.TestCase):
    """Tests for Context methods."""

    def setUp(self):
        """Test harness setup."""
        self.context = nemo.Context()

    def test_total_demand(self):
        """Test total_demand() method."""
        self.assertTrue(self.context.total_demand() > 0)
        self.assertEqual(self.context.total_demand(),
                         self.context.demand.values.sum())

    def test_unserved_energy(self):
        """Test unserved_energy method."""
        self.assertEqual(self.context.unserved_energy(), 0)
        self.assertEqual(self.context.unserved_energy(),
                         self.context.unserved.values.sum())

    def test_surplus_energy(self):
        """Test surplus_energy method."""
        self.assertEqual(self.context.surplus_energy(), 0)
        self.assertEqual(self.context.surplus_energy(),
                         self.context.spill.values.sum())

    def test_unserved_percent(self):
        """Test unserved_percent method."""
        self.assertEqual(self.context.unserved_percent(), 0)

        # Special handling required for zero demand
        self.context.demand = pd.DataFrame()
        self.assertTrue(np.isnan(self.context.unserved_percent()))

    def test_set_capacities(self):
        """Test set_capacities method."""
        self.context.set_capacities([0.1, 0.2])
        self.assertEqual(self.context.generators[0].capacity, 100)
        self.assertEqual(self.context.generators[1].capacity, 200)

    def test_str_no_unserved(self):
        """Test __str__ method (no unserved energy)."""
        output = str(self.context)
        self.assertIn('No unserved energy', output)

    def test_str_with_regions_subset(self):
        """Test __str__ method with only two regions."""
        self.context.regions = [regions.nsw, regions.sa]
        output = str(self.context)
        self.assertIn('Regions: [NSW1, SA1]', output)

    def test_str_no_summary(self):
        """Test __str__ method with a generator that has no summary."""
        self.context.generators[1].summary = lambda _: None
        self.context.verbose = True
        output = str(self.context)
        self.assertIn('OCGT (NSW1:31), 20000.00 MW\nTimesteps:', output)

    def test_str_with_unserved(self):
        """Test __str__ method (with some unserved energy)."""
        self.context.verbose = True
        # Dummy lambda functions for testing
        self.context.surplus_energy = lambda: 300
        self.context.unserved_percent = lambda: 0.5
        rng = pd.date_range(start='2022-01-01', end='2022-01-02', freq='H')
        self.context.unserved = pd.Series(index=rng, data=range(len(rng)))
        output = str(self.context)

        self.assertIn('Generators:', output)
        self.assertIn(f'Timesteps: {self.context.timesteps} h', output)
        self.assertIn('Demand energy:', output)
        self.assertIn('Unstored surplus energy: 300.00 MWh', output)
        self.assertIn('WARNING: reliability standard exceeded', output)
        self.assertIn('Unserved total hours: 25', output)
        self.assertIn('Number of unserved energy events: 1', output)
        self.assertIn('Shortfalls (min, max): (0.00 MW, 24.00 MW)', output)



# Copyright (C) 2017 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""Utility functions (eg, plotting)."""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pint
from matplotlib.patches import Patch
from pandas.plotting import register_matplotlib_converters

from nemo import configfile
from nemo.configfile import configparser

# Future versions of pandas will require us to explicitly register
# matplotlib converters, so do it here now.
register_matplotlib_converters()

# SI units
ureg = pint.UnitRegistry()


def _generator_list(context):
    """Return a list of the generators of interest in this run."""
    return [g for g in context.generators
            if g.region() in context.regions and g.capacity > 0]


def _legend(context):
    """Draw the legend."""
    # ::-1 slicing reverses the list so that the legend appears in merit order
    gens = _generator_list(context)[::-1]
    labels = []
    patches = []

    if len(gens) > 20:
        unique = []
        for gen in gens:
            if type(gen) not in unique:
                unique.append(type(gen))
                labels.append(gen.__class__.__name__)
                patches.append(gen.patch)
    else:
        for gen in gens:
            capacity = (gen.capacity * ureg.MW).to_compact()
            labels.append(gen.label + f' ({capacity:.2f~P})')
            patches.append(gen.patch)

    red_patch = Patch(facecolor='red', edgecolor='black')
    legend = plt.figlegend([red_patch] + patches,
                           ['unserved'] + labels,
                           loc='upper right')
    plt.setp(legend.get_texts(), fontsize='small')


def _figure(context, spills, showlegend, xlim):
    """Provide a helper function for plot() to faciltiate testing."""
    # aggregate demand
    demand = context.demand.sum(axis=1)

    plt.clf()
    plt.ylabel('Power (MW)')
    try:
        title = configfile.get('plot', 'title')
    except (configparser.NoSectionError, configparser.NoOptionError):
        title = 'Energy balance'
    try:
        title += '\n' + configfile.get('plot', 'subtitle')
    except (configfile.configparser.NoSectionError,
            configfile.configparser.NoOptionError):
        pass
    plt.suptitle(title)

    if showlegend:
        _legend(context)

    # Plot demand first.
    plt.plot(demand.index, demand, color='black', linewidth=3 if spills else 2)

    accum = pd.Series(data=0, index=demand.index)
    prev = accum.copy()
    for gen in _generator_list(context):
        idx = context.generators.index(gen)
        accum += context.generation[idx]
        # Ensure accumulated generation does not exceed demand in any timestep.
        # (Due to rounding, accum can be close to demand.)
        assert all(np.logical_or(accum < demand, np.isclose(accum, demand)))
        plt.plot(accum.index, accum, color='black', linewidth=0.4,
                 linestyle='--')
        plt.fill_between(accum.index, prev, accum,
                         facecolor=gen.patch.get_fc(),
                         hatch=gen.patch.get_hatch())
        prev = accum.copy()
    # Unmet demand is shaded red.
    plt.fill_between(accum.index, accum, demand, facecolor='red')

    if spills:
        prev = demand.copy()
        for gen in list(g for g in context.generators if
                        g.region() in context.regions):
            idx = context.generators.index(gen)
            accum += context.spill[idx]
            plt.plot(accum.index, accum, color='black', linewidth=0.4,
                     linestyle='--')
            plt.fill_between(prev.index, prev, accum,
                             facecolor=gen.patch.get_fc(), alpha=0.3)
            prev = accum.copy()

    plt.gca().set_xlim(xlim)  # set_xlim accepts None
    plt.gca().xaxis_date()
    plt.gcf().autofmt_xdate()

    _, ymax = plt.gca().get_ylim()
    plt.plot(context.unserved.index, [ymax] * len(context.unserved),
             "v", markersize=10, color='red', markeredgecolor='black')


def plot(context, spills=False, filename=None, showlegend=True, xlim=None):
    """Produce a pretty plot of supply and demand."""
    _figure(context, spills, showlegend, xlim)
    if not filename:
        plt.show()
    else:
        plt.savefig(filename)



# Copyright (C) 2022 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# pylint: disable=protected-access

"""A testsuite for the sim module."""

import unittest

import numpy as np
import pandas as pd

from nemo import configfile, generators, sim
from nemo.context import Context


class TestSim(unittest.TestCase):
    """Test sim.py."""

    def setUp(self):
        """Test harness setup."""
        self.context = Context()
        self.date_range = pd.date_range('2010-01-01', '2010-01-02', freq='H')
        self.generation = np.zeros((len(self.date_range),
                                    len(self.context.generators)))
        self.spill = np.zeros((len(self.date_range),
                               len(self.context.generators)))

    def test_sim(self):
        """Test _sim() function."""
        self.context.verbose = True
        sim._sim(self.context, self.date_range)

    def test_dispatch(self):
        """Test _dispatch() function."""
        self.context.verbose = True
        sim._dispatch(self.context, 0, 10000, self.context.generators,
                      self.generation, self.spill)
        self.assertEqual(self.spill.sum(), 0)

    def test_dispatch_pv(self):
        """Test _dispatch() function with an async generator."""
        cfg = configfile.get('generation', 'pv1axis-trace')
        pv = generators.PV1Axis(31, 10, cfg, 30)
        # put a 10 MW PV plant at the top of the merit order
        self.context.generators.insert(0, pv)
        self.generation = np.zeros((len(self.date_range),
                                    len(self.context.generators)))
        self.spill = np.zeros((len(self.date_range),
                               len(self.context.generators)))
        sim._dispatch(self.context, 0, 10000, self.context.generators,
                      self.generation, self.spill)
        self.assertEqual(self.spill.sum(), 0)

    def test_store_spills(self):
        """Test _store_spills()."""
        self.context = type('context', (), {'verbose': 0})
        self.context.verbose = True
        hydro = generators.Hydro(1, 100)
        h2store = generators.HydrogenStorage(400)
        electrolyser = generators.Electrolyser(h2store, 1, 100,
                                               efficiency=1.0)
        result = sim._store_spills(self.context, 0, hydro,
                                   [electrolyser], 50)
        self.assertEqual(result, 0.0)
        self.assertEqual(h2store.storage, 250.0)

    def test_store_spills_negative(self):
        """Test _store_spills().

        This test checks proper handling of the case where spl becomes
        very slightly negative. The first pumped hydro generator below
        stores all of the spilled energy, producing a small negative
        residual. The second generator should not be called with zero
        spilled.
        """
        self.context = Context()
        gen = generators.CCGT(31, 200)
        psh1 = generators.PumpedHydro(1, 250, 1000)
        psh1.store = lambda hour, spl: spl + 1e-9
        psh2 = generators.PumpedHydro(1, 250, 1000)
        # this will raise an exception if we try calling store()
        psh2.store = None
        others = [psh1, psh2]
        self.assertEqual(sim._store_spills(self.context, 0, gen,
                                           others, 10), 0)

    def test_run_1(self):
        """Test run() with region not a list."""
        self.context.regions = None
        with self.assertRaises(TypeError):
            sim.run(self.context)

    def test_run_2(self):
        """Test run() normally."""
        sim.run(self.context)



# awklite -- a little library to aid writing AWK-style scripts
#
# Copyright (C) 2022 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""Awklite assists in writing AWK-style scripts in Python.

Awklite is a tiny support library to assist in writing AWK-style
scripts. There are several features:

  * a specialised version of 'list', used for holding fields of a
    line, that can be indexed from one and not zero. In AWK, the first
    field is $1. A simple example:

        >>> f = Fields("1.0 2 three".split())
        >>> f[1], f[2], f[3]
        ('1.0', '2', 'three')

  * a Namespace class that returns an Undefined type (see below) for
    variables that are not attributes of the object. This mimics AWK's
    behaviour of giving undefined variables a null value, eg:

        >>> a = Namespace()
        >>> a.foo = 10
        >>> a.foo
        10
        >>> int(a.bar)
        0

  * a new Undefined type, which mimics an undefined variable in
    AWK. This variable is 0 in an integer context, 0.0 in a float
    context, and the empty string in a string context. As soon as the
    variable is overwritten with a value, it is no longer undefined.
    Some examples to demonstrate the semantics:

        >>> ud = Undefined()
        >>> int(ud), float(ud), str(ud)
        (0, 0.0, '')
        >>> ud + 1, ud + 1.1, ud + 'hello'
        (1, 1.1, 'hello')
        >>> ud > 0
        False
        >>> ud += 1
        >>> ud
        1
"""


class Fields(list):
    """A list which uses one-based indexing."""

    def __getitem__(self, key):
        """Use 1..n list indexing."""
        assert 0 < key <= len(self)
        return list.__getitem__(self, key - 1)


class Undefined():
    """An undefined object mimics an undefined variable in AWK."""

    def __int__(self):
        """Undefined is zero."""
        return 0

    def __float__(self):
        """Undefined is zero."""
        return 0.0

    def __str__(self):
        """Undefined is the empty string."""
        return ""

    def __add__(self, other):
        """X plus Undefined is X."""
        return other

    def __gt__(self, other):
        """Undefined > X is always false."""
        return False


class Namespace():
    """An object that returns an Undefined for undefined attributes.

    >>> ns = Namespace()
    >>> ns.value = 10
    >>> ns.value, type(ns.novalue)
    (10, <class 'awklite.Undefined'>)
    >>> ns.clear()
    >>> type(ns.value)
    <class 'awklite.Undefined'>
    """

    def __getattr__(self, name):
        """Return Undefined for undefined attributes."""
        try:
            return super().__getattr__(self, name)
        except AttributeError:
            return Undefined()

    def clear(self):
        """Clear all variables."""
        self.__dict__.clear()



# Copyright (C) 2022 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""A testsuite for the regions module."""

import copy
import unittest

from nemo import regions


class TestRegions(unittest.TestCase):
    """Tests for Region methods."""

    def test_region(self):
        """Test Region class."""
        rgn = regions.Region(0, 'NSW1', 'New South Wales')
        self.assertEqual('NSW1', str(rgn))
        lst = range(5)
        self.assertEqual(lst[rgn], 0)

    def test_region_copy(self):
        """Check for no copying."""
        sa1 = regions.sa
        sa1copy = copy.copy(sa1)
        self.assertIs(sa1, sa1copy)

    def test_region_deepcopy(self):
        """Check for no deepcopying."""
        sa1 = regions.sa
        sa1copy = copy.deepcopy(sa1)
        self.assertIs(sa1, sa1copy)



# Copyright (C) 2011, 2012, 2014 Ben Elliston
# Copyright (C) 2014, 2015, 2016 The University of New South Wales
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""
Implementation of the Context class.

A simulation context encapsulates all simulation state ensuring that
there is never any residual state left behind after a simulation
run. It also allows multiple contexts to be compared after individual
simulation runs.
"""

import numpy as np
import pandas as pd
import pint

from nemo import configfile, costs, generators, polygons, regions
from nemo.nem import hourly_demand, hourly_regional_demand, startdate

ureg = pint.UnitRegistry()
ureg.default_format = '.2f~P'


class Context():
    """All simulation state is kept in a Context object."""

    # pylint: disable=too-many-instance-attributes
    def __init__(self):
        """Initialise a default context."""
        self.verbose = False
        self.regions = regions.All
        self.startdate = startdate
        # Number of timesteps is determined by the number of demand rows.
        self.hours = len(hourly_regional_demand)
        # Calculate the number of years from the number of simulation hours.
        self.years = self.hours / (365 * 24)

        self.relstd = 0.002  # 0.002% unserved energy
        self.generators = [generators.CCGT(polygons.WILDCARD, 20000),
                           generators.OCGT(polygons.WILDCARD, 20000)]
        self.demand = hourly_demand.copy()
        self.timesteps = len(self.demand)
        self.spill = pd.DataFrame()
        self.generation = pd.DataFrame()
        self.unserved = pd.DataFrame()
        # System non-synchronous penetration limit
        self.nsp_limit = float(configfile.get('limits', 'nonsync-penetration'))
        self.costs = costs.NullCosts()

    def total_demand(self):
        """Return the total demand from the data frame."""
        return self.demand.values.sum()

    def unserved_energy(self):
        """Return the total unserved energy."""
        return self.unserved.values.sum()

    def surplus_energy(self):
        """Return total surplus energy."""
        return self.spill.values.sum()

    def unserved_percent(self):
        """Return the total unserved energy as a percentage of total demand."""
        # We can't catch ZeroDivision because numpy emits a warning
        # (which we would rather not suppress).
        if self.total_demand() == 0:
            return np.nan
        return self.unserved_energy() / self.total_demand() * 100

    def set_capacities(self, caps):
        """Set generator capacities from a list."""
        num = 0
        for gen in self.generators:
            for (setter, min_cap, max_cap) in gen.setters:
                # keep parameters within bounds
                newval = max(min(caps[num], max_cap), min_cap)
                setter(newval)
                num += 1
        # Check every parameter has been set.
        assert num == len(caps), f'{num} != {len(caps)}'

    def __str__(self):
        """Make a human-readable representation of the context."""
        string = ""
        if self.regions != regions.All:
            string += f'Regions: {self.regions}\n'
        if self.verbose:
            string += 'Generators:' + '\n'
            for gen in self.generators:
                string += f'\t{gen}'
                summary = gen.summary(self)
                if summary is not None:
                    string += f'\n\t   {summary}\n'
                else:
                    string += '\n'
        string += f'Timesteps: {self.hours} h\n'
        total_demand = (self.total_demand() * ureg.MWh).to_compact()
        string += f'Demand energy: {total_demand}\n'
        surplus_energy = (self.surplus_energy() * ureg.MWh).to_compact()
        string += f'Unstored surplus energy: {surplus_energy}\n'
        if self.surplus_energy() > 0:
            spill_series = self.spill[self.spill.sum(axis=1) > 0]
            string += 'Timesteps with unused surplus energy: '
            string += f'{len(spill_series)}\n'

        if self.unserved.empty:
            string += 'No unserved energy'
        else:
            string += f'Unserved energy: {self.unserved_percent():.3f}%\n'
            if self.unserved_percent() > self.relstd * 1.001:
                string += 'WARNING: reliability standard exceeded\n'
            string += f'Unserved total hours: {len(self.unserved)}\n'

            # A subtle trick: generate a date range and then subtract
            # it from the timestamps of unserved events.  This will
            # produce a run of time deltas (for each consecutive hour,
            # the time delta between this timestamp and the
            # corresponding row from the range will be
            # constant). Group by the deltas.
            date_range = pd.date_range(self.unserved.index[0],
                                       periods=len(self.unserved.index),
                                       freq='H')
            deltas = self.unserved.groupby(self.unserved.index - date_range)
            unserved_events = [k for k, g in deltas]
            string += 'Number of unserved energy events: '
            string += f'{len(unserved_events)}\n'
            if not self.unserved.empty:
                umin = (self.unserved.min() * ureg.MW).to_compact()
                umax = (self.unserved.max() * ureg.MW).to_compact()
                string += f'Shortfalls (min, max): ({umin}, {umax})'
        return string



# Copyright (C) 2021 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""Penalty functions for the optimisation."""

from nemo import generators

_reason_labels = ['unserved', 'emissions', 'fossil', 'bioenergy',
                  'hydro', 'reserves', 'min-regional-gen']

reasons = {}
for i, label in enumerate(_reason_labels):
    reasons[label] = 1 << i

# Conversion factor between MWh and TWh.
_twh = pow(10., 6)


def unserved(ctx, _):
    """Penalty: unserved energy."""
    minuse = ctx.total_demand() * (ctx.relstd / 100)
    use = max(0, ctx.unserved_energy() - minuse)
    reason = reasons['unserved'] if use > 0 else 0
    return pow(use, 3), reason


def _calculate_reserve(gen, time):
    """Calculate headroom for each generator.

    Note: except pumped hydro and CST -- tricky to calculate capacity.
    """
    if isinstance(gen, generators.Fuelled) and not \
       isinstance(gen, generators.PumpedHydro) and not \
       isinstance(gen, generators.CST):
        return gen.capacity - gen.series_power[time]
    return 0


def reserves(ctx, args):
    """Penalty: minimum reserves."""
    pen, reas = 0, 0
    for time in range(ctx.timesteps):
        reserve, spilled = 0, 0
        for gen in ctx.generators:
            try:
                spilled += gen.series_spilled[time]
            except KeyError:
                # non-variable generators may not have spill data
                pass
            reserve += _calculate_reserve(gen, time)

        if reserve + spilled < args.reserves:
            reas |= reasons['reserves']
            pen += pow(args.reserves - reserve + spilled, 3)
    return pen, reas


def _regional_generation(region, gens):
    """Sum generation in a given region."""
    regional_generation = 0
    for gen in gens:
        if gen.region() is region:
            regional_generation += sum(gen.series_power.values())
    return regional_generation


def _regional_demand(region, demand):
    """Sum demand in a given region."""
    regional_demand = 0
    for poly in region.polygons:
        regional_demand += demand[poly - 1].sum()
    return regional_demand


def min_regional(ctx, _):
    """Penalty: minimum share of regional generation."""
    shortfall = 0
    for rgn in ctx.regions:
        regional_demand = _regional_demand(rgn, ctx.demand)
        regional_generation = _regional_generation(rgn, ctx.generators)
        min_regional_generation = regional_demand * ctx.min_regional_generation
        shortfall += max(0, min_regional_generation - regional_generation)

    if shortfall > 0:
        reason = reasons['min-regional-gen']
    else:
        reason = 0
    return pow(shortfall, 3), reason


def emissions(ctx, args):
    """Penalty: total emissions."""
    total_emissions = 0
    for gen in ctx.generators:
        if hasattr(gen, 'intensity'):
            total_emissions += sum(gen.series_power.values()) * gen.intensity
    emissions_limit = args.emissions_limit * pow(10, 6) * ctx.years
    # exceedance in tonnes CO2-e
    emissions_exceedance = max(0, total_emissions - emissions_limit)
    reason = reasons['emissions'] if emissions_exceedance > 0 else 0
    return pow(emissions_exceedance, 3), reason


def fossil(ctx, args):
    """Penalty: limit fossil to fraction of annual demand."""
    fossil_energy = 0
    for gen in ctx.generators:
        if isinstance(gen, generators.Fossil):
            fossil_energy += sum(gen.series_power.values())
    fossil_limit = ctx.total_demand() * args.fossil_limit * ctx.years
    fossil_exceedance = max(0, fossil_energy - fossil_limit)
    reason = reasons['fossil'] if fossil_exceedance > 0 else 0
    return pow(fossil_exceedance, 3), reason


def bioenergy(ctx, args):
    """Penalty: limit biofuel use."""
    biofuel_energy = 0
    for gen in ctx.generators:
        if isinstance(gen, generators.Biofuel):
            biofuel_energy += sum(gen.series_power.values())
    biofuel_limit = args.bioenergy_limit * _twh * ctx.years
    biofuel_exceedance = max(0, biofuel_energy - biofuel_limit)
    reason = reasons['bioenergy'] if biofuel_exceedance > 0 else 0
    return pow(biofuel_exceedance, 3), reason


def hydro(ctx, args):
    """Penalty: limit hydro use."""
    hydro_energy = 0
    for gen in ctx.generators:
        if isinstance(gen, generators.Hydro) and \
           not isinstance(gen, generators.PumpedHydro):
            hydro_energy += sum(gen.series_power.values())
    hydro_limit = args.hydro_limit * _twh * ctx.years
    hydro_exceedance = max(0, hydro_energy - hydro_limit)
    reason = reasons['hydro'] if hydro_exceedance > 0 else 0
    return pow(hydro_exceedance, 3), reason



# Copyright (C) 2022 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# pylint: disable=protected-access

"""A testsuite for the penalties module."""

import unittest

import numpy as np

import nemo
from nemo import generators, penalties, regions
from nemo.penalties import reasons
from nemo.polygons import WILDCARD


class Args:
    """Faked up command line options."""

    emissions_limit = 0
    fossil_limit = 0.5  # 50%
    bioenergy_limit = 1e-6  # 1 MWh
    hydro_limit = 1e-6  # 1 MWh
    reserves = 50  # MW


args = Args()


class TestPenalties(unittest.TestCase):
    """Test functions in penalties.py."""

    def setUp(self):
        """Test harness setup."""
        self.context = nemo.Context()
        # Override standard attributes and methods for testing
        self.context.demand = np.ones((43, 100))
        self.context.total_demand = lambda: 100
        self.context.unserved_energy = lambda: 0.01
        self.context.relstd = 0

    def test_unserved(self):
        """Test unserved() function."""
        self.assertEqual(penalties.unserved(self.context, 0),
                         (pow(0.01, 3), reasons['unserved']))

    def test_calculate_reserve(self):
        """Test _calculate_reserve() function."""
        self.context.generators[0].series_power = {n: 1 for n in range(1000)}
        generator = self.context.generators[0]
        capacity = generator.capacity
        self.assertEqual(penalties._calculate_reserve(generator, 0),
                         capacity - 1)
        psh = generators.PumpedHydro(WILDCARD, 0, 0)
        self.assertEqual(penalties._calculate_reserve(psh, 0), 0)

    def test_reserves(self):
        """Test reserves() function."""
        self.context.timesteps = 100
        del self.context.generators[1:]
        # 55 MW x 100 hours, 5 MW over reserve level
        self.context.generators[0].series_power = {n: 55 for n in range(100)}
        self.context.generators[0].capacity = 100
        self.assertEqual(penalties.reserves(self.context, args),
                         (pow(5, 3) * 100, reasons['reserves']))

    def test_regional_generation(self):
        """Test _regional_generation() function."""
        # Gen 1: 1,000 MWh, Gen 2: 1,000 MWh, total 2,000 MWh
        self.context.generators[0].series_power = {n: 1 for n in range(1000)}
        self.context.generators[1].series_power = {n: 1 for n in range(1000)}
        # both generators are in NSW
        self.assertEqual(
            penalties._regional_generation(regions.nsw,
                                           self.context.generators), 2000)
        self.assertEqual(
            penalties._regional_generation(regions.sa,
                                           self.context.generators), 0)

    def test_regional_demand(self):
        """Test _regional_demand() function."""
        for rgn in regions.All:
            # Check that there is X00 MWh of demand per region,
            # 100 MWh per polygon
            polycount = len(rgn.polygons) * 100
            self.assertEqual(penalties._regional_demand(rgn,
                                                        self.context.demand),
                             polycount)

    def test_min_regional_0(self):
        """Test min_regional() function at 0%."""
        self.context.min_regional_generation = 0
        self.assertEqual(penalties.min_regional(self.context, args), (0, 0))

    def test_min_regional_50(self):
        """Test min_regional() function at 50%."""
        self.context.min_regional_generation = 0.5
        # just two regions: NSW and SA
        self.context.regions = [regions.nsw, regions.sa]
        self.assertEqual(penalties.min_regional(self.context, args),
                         (pow(1050, 3), reasons['min-regional-gen']))

    def test_emissions(self):
        """Test emissions() function."""
        # Gen 1: 1,000 MWh (1 GWh) at 0.8 tonnes/MWh = 800 t
        # Gen 2: 1,000 MWh (1 GWh) at 0.5 tonnes/MWh = 500 t
        # Total: 1,300 tonnes
        self.context.generators[0].series_power = {n: 1 for n in range(1000)}
        self.context.generators[0].intensity = 0.800
        self.context.generators[1].series_power = {n: 1 for n in range(1000)}
        self.context.generators[1].intensity = 0.500

        self.assertEqual(penalties.emissions(self.context, args),
                         (pow(1300, 3), reasons['emissions']))

    def test_fossil(self):
        """Test fossil() function."""
        # Gen 1: 10 MWh, Gen 2: 10 MWh (Total 20MWh or 20% of demand)
        self.context.generators[0].series_power = {n: 1 for n in range(10)}
        self.context.generators[1].series_power = {n: 1 for n in range(10)}
        self.assertEqual(penalties.fossil(self.context, args), (0, 0))

        # Gen 1: 50 MWh, Gen 2: 50 MWh (Total 100MWh or 100% of demand)
        self.assertEqual(args.fossil_limit, 0.5)
        self.context.generators[0].series_power = {n: 5 for n in range(10)}
        self.context.generators[1].series_power = {n: 5 for n in range(10)}
        self.assertEqual(penalties.fossil(self.context, args),
                         (pow(50, 3), reasons['fossil']))

    def test_bioenergy(self):
        """Test bioenerge() function."""
        bio = generators.Biofuel(WILDCARD, 0)
        self.context.generators += [bio]
        # bioenergy: 0 MWh
        bio.series_power = {}
        self.assertEqual(penalties.bioenergy(self.context, args), (0, 0))
        # bioenergy: 5 MWh
        bio.series_power = {n: 1 for n in range(5)}
        self.assertEqual(penalties.bioenergy(self.context, args),
                         (pow(4, 3), reasons['bioenergy']))

    def test_hydro(self):
        """Test hydro() function."""
        hydro = generators.Hydro(WILDCARD, 0)
        self.context.generators += [hydro]
        # hydro: 0 MWh
        hydro.series_power = {}
        self.assertEqual(penalties.hydro(self.context, args), (0, 0))
        # hydro: 5 MWh
        hydro.series_power = {n: 1 for n in range(5)}
        self.assertEqual(penalties.hydro(self.context, args),
                         (pow(4, 3), reasons['hydro']))



# Copyright (C) 2022 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""Useful internal types."""


class UnreachableError(AssertionError):
    """For marking unreachable code."""



# Copyright (C) 2014 The University of New South Wales
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""Configuration file processing (eg, filenames)."""

import configparser
import os


def load(filename):
    """Load a configuration file (or files)."""
    result = config.read(filename)
    if not result:
        raise FileNotFoundError(f"config file {filename} not found")
    # Verify
    config.get('generation', 'cst-trace')
    config.get('generation', 'egs-geothermal-trace')
    config.get('generation', 'hsa-geothermal-trace')
    config.get('generation', 'wind-trace')
    config.get('generation', 'pv1axis-trace')
    config.get('demand', 'demand-trace')


def get(section, option):
    """
    Get an option value for the named section.

    This works the same as ConfigParser.get.
    """
    return config.get(section, option)


def has_option_p(section, option):
    """
    Check if this section has a given option.

    This works the same as ConfigParser.has_option.
    """
    return config.has_option(section, option)


config = configparser.ConfigParser()

# If $NEMORC is set, use that as the config filename.
if os.getenv('NEMORC') is not None:
    load(os.getenv('NEMORC'))
else:
    load('nemo.cfg')



# Copyright (C) 2011, 2014 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""A testsuite for NEMO."""

import math
import unittest

import nemo
from nemo import generators, polygons, regions


class SuperGenerator(generators.Generator):
    """A synthetic generator that can always meet demand."""

    def __init__(self, capacity):
        """Create a super generator."""
        generators.Generator.__init__(self, polygons.WILDCARD, capacity)
        self.energy = 0
        self.runhours = 0

    def reset(self):
        """Reset context between tests."""
        self.energy = 0
        self.runhours = 0

    def step(self, _, demand):
        """Step the super generator."""
        self.runhours += 1
        self.energy += demand
        if self.capacity == 0:
            # meet demand exactly
            return demand, 0
        # meet demand, spill surplus capacity
        surplus = max(0, self.capacity - demand)
        return demand, surplus


class TestSequenceFunctions(unittest.TestCase):
    """Basic tests for now."""

    def setUp(self):
        """Test harness setup."""
        self.context = nemo.Context()
        self.minload = math.floor(self.context.demand.sum(axis=1).min())

    def test_001(self):
        """Test that all regions are present."""
        self.assertEqual(self.context.regions, regions.All)

    def test_002(self):
        """Demand equals approx. 204 TWh."""
        self.context.generators = []
        nemo.run(self.context)
        total_demand = math.trunc(self.context.total_demand() / pow(10., 6))
        self.assertEqual(total_demand, 204)

    def test_003(self):
        """Power system with no generators meets none of the demand."""
        self.context.generators = []
        nemo.run(self.context)
        self.assertEqual(math.trunc(self.context.unserved_energy()),
                         math.trunc(self.context.total_demand()))

    def test_004(self):
        """100 MW fossil plant generates exactly 876,000 MWh."""
        ccgt = generators.CCGT(polygons.WILDCARD, 100)
        self.context.generators = [ccgt]
        nemo.run(self.context)
        total_generation = sum(ccgt.series_power.values())
        expected_generation = self.context.timesteps * 100
        self.assertEqual(total_generation, expected_generation)

    # Create a super generator that always meets demand.
    # Check unserved_energy = 0

    def test_005(self):
        """Super generator runs every hour."""
        gen = SuperGenerator(0)
        self.context.generators = [gen]
        nemo.run(self.context)
        self.assertEqual(gen.runhours, self.context.timesteps)

    def test_006(self):
        """Generation to meet minimum load leads to no spills."""
        self.context.generators = [SuperGenerator(self.minload)]
        nemo.run(self.context)
        self.assertEqual(self.context.spill.values.sum(), 0)

    def test_007(self):
        """Generation to meet minimum load + 1GW produces some spills."""
        self.context.generators = [SuperGenerator(self.minload + 1000)]
        nemo.run(self.context)
        self.assertTrue(self.context.spill.values.sum() > 0)

    def test_008(self):
        """A NSW generator runs in NSW only."""
        for rgn in regions.All:
            self.context.regions = [rgn]
            gen = SuperGenerator(0)
            self.context.generators = [gen]
            nemo.run(self.context)
            if rgn == regions.nsw:
                self.assertEqual(gen.runhours, self.context.timesteps)
            else:
                self.assertEqual(gen.runhours, 0)

    def test_009(self):
        """A NSW generators runs in any set of regions that includes NSW."""
        rgnset = []
        for rgn in regions.All:
            rgnset.append(rgn)
            self.context.regions = rgnset
            gen = SuperGenerator(0)
            self.context.generators = [gen]
            nemo.run(self.context)
            self.assertEqual(gen.runhours, self.context.timesteps)

    def test_012(self):
        """A NSW generator does not run in other regions."""
        rgnset = []
        # Skip NSW (first in the list).
        for rgn in regions.All[1:]:
            rgnset.append(rgn)
            self.context.regions = rgnset
            gen = SuperGenerator(0)
            self.context.generators = [gen]
            nemo.run(self.context)
            self.assertEqual(gen.runhours, 0)

    def test_013(self):
        """Fossil plant records power generation history."""
        ccgt = generators.CCGT(polygons.WILDCARD, 100)
        self.context.generators = [ccgt]
        nemo.run(self.context)
        self.assertTrue(len(self.context.generators[0].series_power) > 0)



# Copyright (C) 2011, 2012, 2014 Ben Elliston
# Copyright (C) 2014, 2015, 2016 The University of New South Wales
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""A National Electricity Market (NEM) simulation."""

import io

import numpy as np
import pandas as pd
import requests

from nemo import configfile, polygons, regions

# Demand is in 30 minute intervals. NOTE: the number of rows in the
# demand file now dictates the number of timesteps in the simulation.

url = configfile.get('demand', 'demand-trace')

if not url.startswith('http'):
    # Local file path
    traceinput = url
else:
    try:
        resp = requests.request('GET', url, timeout=5)
    except requests.exceptions.Timeout as exc:
        raise TimeoutError(f'timeout fetching {url}') from exc
    if not resp.ok:
        raise ConnectionError(f'HTTP {resp.status_code}: {url}')
    traceinput = io.StringIO(resp.text)

demand = pd.read_csv(traceinput, comment='#', sep=',',
                     parse_dates=[['Date', 'Time']], index_col='Date_Time')

# Check for date, time and n demand columns (for n regions).
assert len(demand.columns) == regions.NUMREGIONS
# The number of rows must be even.
assert len(demand) % 2 == 0, "odd number of rows in half-hourly demand data"

# Check demand data starts at midnight
startdate = demand.index[0]
assert (startdate.hour, startdate.minute, startdate.second) == (0, 30, 0), \
    'demand data must start at midnight'

# Calculate hourly demand, averaging half-hours n and n+1.
hourly_regional_demand = demand.resample('H', closed='right').mean()

# Now put the demand into polygon resolution according to the load
# apportioning figures given in each region's polygons field.
numsteps = len(hourly_regional_demand)
hourly_demand = pd.DataFrame(index=hourly_regional_demand.index,
                             data=np.zeros((numsteps, polygons.NUMPOLYGONS)))

for rgn, weights in [(r.id, r.polygons) for r in regions.All]:
    for polygon, share in weights.items():
        hourly_demand[polygon - 1] = hourly_regional_demand[rgn] * share



# Copyright (C) 2012, 2013 Ben Elliston
# Copyright (C) 2014, 2015 The University of New South Wales
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# We use class names here that upset Pylint.
# pylint: disable=invalid-name
# pylint: disable=too-many-instance-attributes

"""Generation technology costs."""
from nemo import generators as tech


def annuity_factor(lifetime, rate):
    """Return the annuity factor for lifetime t and interest rate r."""
    return (1 - (1 / (1 + rate) ** lifetime)) / rate


class NullCosts():
    """All costs are zero. Useful for debugging."""

    class _ZeroDict(dict):
        """Return a fixed value (eg, 0) for any key."""

        def __init__(self, value=0):
            dict.__init__(self)
            self.value = value

        def __getitem__(self, key):
            return dict.get(self, key, self.value)

    # pylint: disable=unused-argument
    def __init__(self, discount=0, coal_price=0, gas_price=0, ccs_price=0):
        """Construct an all-zero costs object."""
        self.capcost_per_kw = self._ZeroDict()
        self.fixed_om_costs = self._ZeroDict()
        self.opcost_per_mwh = self._ZeroDict()
        # a dictionary of dictionary of zeros
        self.totcost_per_kwh = self._ZeroDict(self._ZeroDict())
        self.annuityf = 1
        self.ccs_storage_per_t = 0
        self.bioenergy_price_per_gj = 0
        self.coal_price_per_gj = 0
        self.gas_price_per_gj = 0
        self.diesel_price_per_litre = 0
        self.carbon = 0


class APGTR2015():
    """Australian Power Generation Technology Report costs in 2015.

    Source: CO2CRC Australian Power Generation Technology Report (2015)
    """

    lifetime = 30
    escalation = 1.0

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        self.discount_rate = discount
        self.ccs_storage_per_t = ccs_price
        # bioenergy costs taken from CSIRO energy storage report for AEMO
        self.bioenergy_price_per_gj = 12
        self.coal_price_per_gj = coal_price
        self.gas_price_per_gj = gas_price
        self.diesel_price_per_litre = 1.50
        self.annuityf = annuity_factor(self.lifetime, discount)

        # Common capital costs
        self.capcost_per_kw = {
            tech.Hydro: 0,
            tech.PumpedHydro: 0,
            tech.Diesel: 0,
            tech.DemandResponse: 0}

        # Variable O&M (VOM) costs
        self.opcost_per_mwh = {
            tech.Hydro: 0,
            tech.PumpedHydro: 0,
            tech.Diesel: 0,
            tech.Wind: 0,
            tech.CentralReceiver: 4,
            tech.PV: 0,
            tech.PV1Axis: 0,
            tech.CCGT: 1.5,
            tech.OCGT: 12,
            tech.Black_Coal: 2.5}

        # Fixed O&M (FOM) costs
        self.fixed_om_costs = {
            tech.DemandResponse: 0,
            tech.Diesel: 0,
            tech.Hydro: 0,
            tech.PumpedHydro: 0,
            tech.Wind: 55,
            tech.CentralReceiver: 65,
            tech.PV: 30,
            tech.PV1Axis: 35,
            tech.CCGT: 20,
            tech.OCGT: 8,
            tech.Black_Coal: 45}

        table = self.capcost_per_kw
        table[tech.Wind] = 2450
        table[tech.CentralReceiver] = 8500
        table[tech.PV] = 2100
        table[tech.PV1Axis] = 2700
        table[tech.CCGT] = 1450
        table[tech.OCGT] = 1000
        table[tech.Black_Coal] = 3000


class APGTR2030(APGTR2015):
    """Australian Power Generation Technology Report (2015) costs in 2030.

    Source: CO2CRC Australian Power Generation Technology Report (2015)
    """

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        APGTR2015.__init__(self, discount, coal_price, gas_price, ccs_price)

        # Modify the capital costs in APGTR2015 by specified learning rates.
        # Fixed and variable O&M remain the same as in 2015.
        table = self.capcost_per_kw
        table[tech.Wind] *= 0.8
        table[tech.CentralReceiver] *= 0.8
        table[tech.PV] *= 0.5
        table[tech.PV1Axis] *= 0.5
        table[tech.CCGT] *= 0.9
        table[tech.OCGT] *= 1.1
        table[tech.Black_Coal] *= 0.9


class AETA2012_2030():
    """Australian Energy Technology Assessment (2012) costs for 2030.

    Source: BREE AETA report (2012), bree.gov.au
    """

    lifetime = 30
    escalation = 1.171

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        self.discount_rate = discount
        self.ccs_storage_per_t = ccs_price
        # bioenergy costs taken from CSIRO energy storage report for AEMO
        self.bioenergy_price_per_gj = 12
        self.coal_price_per_gj = coal_price
        self.gas_price_per_gj = gas_price
        self.diesel_price_per_litre = 1.50
        self.annuityf = annuity_factor(self.lifetime, discount)

        # Common capital costs
        self.capcost_per_kw = {
            tech.Hydro: 0,
            tech.PumpedHydro: 0,
            tech.Diesel: 0,
            tech.DemandResponse: 0}

        # Variable O&M (VOM) costs
        self.opcost_per_mwh = {
            tech.Hydro: 0,
            tech.PumpedHydro: 0,
            tech.Diesel: 0,
            tech.Wind: 12 * self.escalation,
            tech.CentralReceiver: 15 * self.escalation,
            tech.ParabolicTrough: 20 * self.escalation,
            tech.PV: 0,
            tech.PV1Axis: 0,
            tech.CCGT: 4 * self.escalation,
            tech.OCGT: 10 * self.escalation,
            tech.CCGT_CCS: 9 * self.escalation,
            tech.Coal_CCS: 15 * self.escalation,
            tech.Black_Coal: 7 * self.escalation,
            tech.Geothermal_HSA: 0,
            tech.Geothermal_EGS: 0}

        # Fixed O&M (FOM) costs
        self.fixed_om_costs = {
            tech.DemandResponse: 0,
            tech.Diesel: 0,
            tech.Hydro: 0,
            tech.PumpedHydro: 0,
            tech.Wind: 40 * self.escalation,
            tech.CentralReceiver: 60 * self.escalation,
            tech.ParabolicTrough: 65 * self.escalation,
            tech.PV: 25 * self.escalation,
            tech.PV1Axis: 38 * self.escalation,
            tech.CCGT: 10 * self.escalation,
            tech.OCGT: 4 * self.escalation,
            tech.CCGT_CCS: 17 * self.escalation,
            tech.Coal_CCS: 73.2 * self.escalation,
            tech.Black_Coal: 50.5 * self.escalation,
            tech.Geothermal_HSA: 200 * self.escalation,
            tech.Geothermal_EGS: 170 * self.escalation}


class AETA2012_2030Low(AETA2012_2030):
    """AETA (2012) costs for 2030, low end of the range."""

    def __init__(self, discount, coal_price, gas_price, ccs_storage_costs):
        """Construct a cost object."""
        AETA2012_2030.__init__(self, discount, coal_price, gas_price,
                               ccs_storage_costs)
        # capital costs in $/kW
        table = self.capcost_per_kw
        table[tech.Wind] = 1701
        table[tech.CentralReceiver] = 4203
        table[tech.ParabolicTrough] = 4563
        table[tech.PV] = 1482
        table[tech.PV1Axis] = 2013
        table[tech.CCGT] = 1015
        table[tech.OCGT] = 694
        table[tech.CCGT_CCS] = 2095
        table[tech.Coal_CCS] = 4453
        table[tech.Black_Coal] = 2947
        table[tech.Geothermal_HSA] = 6645
        table[tech.Geothermal_EGS] = 10331


class AETA2012_2030High(AETA2012_2030):
    """AETA (2012) costs for 2030, high end of the range."""

    def __init__(self, discount, coal_price, gas_price, ccs_storage_costs):
        """Construct a cost object."""
        AETA2012_2030.__init__(self, discount, coal_price, gas_price,
                               ccs_storage_costs)
        # capital costs in $/kW
        table = self.capcost_per_kw
        table[tech.Wind] = 1917
        table[tech.CentralReceiver] = 5253
        table[tech.ParabolicTrough] = 5659
        table[tech.PV] = 1871
        table[tech.PV1Axis] = 2542
        table[tech.CCGT] = 1221
        table[tech.OCGT] = 809
        table[tech.CCGT_CCS] = 2405
        table[tech.Coal_CCS] = 4727
        table[tech.Black_Coal] = 3128
        table[tech.Geothermal_HSA] = 7822
        table[tech.Geothermal_EGS] = 11811


class AETA2012_2030Mid(AETA2012_2030):
    """AETA (2012) costs for 2030, middle of the range."""

    def __init__(self, discount, coal_price, gas_price, ccs_storage_costs):
        """Construct a cost object."""
        AETA2012_2030.__init__(self, discount, coal_price, gas_price,
                               ccs_storage_costs)

        low = AETA2012_2030Low(discount, coal_price, gas_price,
                               ccs_storage_costs)
        high = AETA2012_2030High(discount, coal_price, gas_price,
                                 ccs_storage_costs)
        assert low.opcost_per_mwh == high.opcost_per_mwh
        assert low.fixed_om_costs == high.fixed_om_costs

        table = self.capcost_per_kw
        lowtable = low.capcost_per_kw
        hightable = high.capcost_per_kw
        for key, lowcost in lowtable.items():
            highcost = hightable[key]
            table[key] = lowcost / 2 + highcost / 2


class AETA2013_2030Low(AETA2012_2030Low):
    """AETA (2013 update) costs for 2030, low end of the range."""

    def __init__(self, discount, coal_price, gas_price, ccs_storage_costs):
        """Construct a cost object."""
        AETA2012_2030Low.__init__(self, discount, coal_price, gas_price,
                                  ccs_storage_costs)

        # Override a few O&M costs.
        fom = self.fixed_om_costs
        fom[tech.Wind] = 32.5 * self.escalation
        fom[tech.PV1Axis] = 30 * self.escalation
        fom[tech.CentralReceiver] = 71.312 * self.escalation
        fom[tech.ParabolicTrough] = 72.381 * self.escalation
        vom = self.opcost_per_mwh
        vom[tech.Wind] = 10 * self.escalation
        vom[tech.CentralReceiver] = 5.65 * self.escalation
        vom[tech.ParabolicTrough] = 11.39 * self.escalation


class AETA2013_2030High(AETA2012_2030High):
    """AETA (2013 update) costs for 2030, high end of the range."""

    def __init__(self, discount, coal_price, gas_price, ccs_storage_costs):
        """Construct a cost object."""
        AETA2012_2030High.__init__(self, discount, coal_price, gas_price,
                                   ccs_storage_costs)

        # Override a few O&M costs.
        fom = self.fixed_om_costs
        fom[tech.Wind] = 32.5 * self.escalation
        fom[tech.PV1Axis] = 30 * self.escalation
        fom[tech.CentralReceiver] = 71.312 * self.escalation
        fom[tech.ParabolicTrough] = 72.381 * self.escalation
        vom = self.opcost_per_mwh
        vom[tech.Wind] = 10 * self.escalation
        vom[tech.CentralReceiver] = 5.65 * self.escalation
        vom[tech.ParabolicTrough] = 11.39 * self.escalation


class AETA2013_2030Mid(AETA2012_2030):
    """AETA (2013) costs for 2030, middle of the range."""

    def __init__(self, discount, coal_price, gas_price, ccs_storage_costs):
        """Construct a cost object."""
        AETA2012_2030.__init__(self, discount, coal_price, gas_price,
                               ccs_storage_costs)

        low = AETA2013_2030Low(discount, coal_price, gas_price,
                               ccs_storage_costs)
        high = AETA2013_2030High(discount, coal_price, gas_price,
                                 ccs_storage_costs)
        assert low.opcost_per_mwh == high.opcost_per_mwh
        self.opcost_per_mwh = low.opcost_per_mwh
        assert low.fixed_om_costs == high.fixed_om_costs
        self.fixed_om_costs = low.fixed_om_costs

        table = self.capcost_per_kw
        lowtable = low.capcost_per_kw
        hightable = high.capcost_per_kw
        for key, lowcost in lowtable.items():
            highcost = hightable[key]
            table[key] = lowcost / 2 + highcost / 2


class CEEM2016_2030(AETA2012_2030Mid):
    """
    CEEM 2016 custom costs.

    These custom costs were produced by CEEM -- AETA (2013) mid costs
    with CO2CRC Power Generation Technology Report 2030 capital costs
    for utility-scale PV.
    """

    def __init__(self, discount, coal_price, gas_price, ccs_storage_costs):
        """Construct a cost object."""
        AETA2012_2030Mid.__init__(self, discount, coal_price, gas_price,
                                  ccs_storage_costs)

        # CO2CRC Power Generation Technology Report (p. 253) gives a
        # narrow range of $1,108 to $1,218 per kW. Meet half-way.
        self.capcost_per_kw[tech.PV1Axis] = 1255


class GenCost2021:
    """GenCost 2020-21 costs.

    Source:
    CSIRO GenCost 2020-21 report
    https://data.csiro.au/collections/collection/CIcsiro:44228
    """

    lifetime = 30

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        self.discount_rate = discount
        self.ccs_storage_per_t = ccs_price
        # bioenergy costs taken from CSIRO energy storage report for AEMO
        self.bioenergy_price_per_gj = 12
        self.coal_price_per_gj = coal_price
        self.gas_price_per_gj = gas_price
        self.diesel_price_per_litre = 1.50
        self.annuityf = annuity_factor(self.lifetime, discount)

        # Fixed O&M (FOM) costs
        # Note: These are the same for all years (2030, 2040, 2050),
        # so we can set them once here.
        self.fixed_om_costs = {
            tech.Black_Coal: 53.2,
            tech.CCGT: 10.9,
            tech.CCGT_CCS: 16.4,
            tech.CentralReceiver: 142.5,
            tech.Coal_CCS: 77.8,
            tech.Hydro: 0,
            tech.OCGT: 10.2,
            tech.PV1Axis: 17.0,
            tech.PumpedHydro: 0,
            tech.Wind: 25.0,
            tech.DemandResponse: 0
        }

        # Variable O&M (VOM) costs
        # Likewise, these are the same for all years (2030, 2040, 2050).
        self.opcost_per_mwh = {
            tech.Black_Coal: 4.2,
            tech.CCGT: 3.7,
            tech.CCGT_CCS: 7.2,
            tech.CentralReceiver: 0,
            tech.Coal_CCS: 8.0,
            tech.Hydro: 0,
            tech.OCGT: 2.4,
            tech.PV1Axis: 0,
            tech.PumpedHydro: 0,
            tech.Wind: 0
        }

        # Common capital costs
        self.capcost_per_kw = {
            tech.Hydro: 0,
            tech.PumpedHydro: 0,
            tech.DemandResponse: 0
        }


class GenCost2021_2020(GenCost2021):
    """GenCost 2020-21 costs for 2020."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2021.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 4450
        table[tech.CCGT] = 1743
        table[tech.CCGT_CCS] = 4396
        table[tech.CentralReceiver] = 7411
        table[tech.Coal_CCS] = 9311
        table[tech.OCGT] = 873
        table[tech.PV1Axis] = 1505
        table[tech.Wind] = 1951


class GenCost2021_2030Low(GenCost2021):
    """GenCost 2020-21 costs for 2030 (low end of the range)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2021.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 4362
        table[tech.CCGT] = 1709
        table[tech.CCGT_CCS] = 3865
        table[tech.CentralReceiver] = 5968
        table[tech.Coal_CCS] = 8674
        table[tech.OCGT] = 856
        table[tech.PV1Axis] = 768
        table[tech.Wind] = 1863


class GenCost2021_2030High(GenCost2021):
    """GenCost 2020-21 costs for 2030 (high end of the range)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2021.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 4362
        table[tech.CCGT] = 1709
        table[tech.CCGT_CCS] = 4352
        table[tech.CentralReceiver] = 6496
        table[tech.Coal_CCS] = 14054
        table[tech.OCGT] = 856
        table[tech.PV1Axis] = 933
        table[tech.Wind] = 1910


class GenCost2021_2040Low(GenCost2021):
    """GenCost 2020-21 costs for 2040 (low end of the range)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2021.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 4275
        table[tech.CCGT] = 1675
        table[tech.CCGT_CCS] = 3327
        table[tech.CentralReceiver] = 5234
        table[tech.Coal_CCS] = 8030
        table[tech.OCGT] = 839
        table[tech.PV1Axis] = 569
        table[tech.Wind] = 1822


class GenCost2021_2040High(GenCost2021):
    """GenCost 2020-21 costs for 2040 (high end of the range)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2021.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 4275
        table[tech.CCGT] = 1675
        table[tech.CCGT_CCS] = 4309
        table[tech.CentralReceiver] = 6087
        table[tech.Coal_CCS] = 9034
        table[tech.OCGT] = 839
        table[tech.PV1Axis] = 778
        table[tech.Wind] = 1863


class GenCost2021_2050Low(GenCost2021):
    """GenCost 2020-21 costs for 2050 (low end of the range)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2021.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 4195
        table[tech.CCGT] = 1643
        table[tech.CCGT_CCS] = 3276
        table[tech.CentralReceiver] = 4748
        table[tech.Coal_CCS] = 7891
        table[tech.OCGT] = 822
        table[tech.PV1Axis] = 532
        table[tech.Wind] = 1774


class GenCost2021_2050High(GenCost2021):
    """GenCost 2020-21 costs for 2050 (high end of the range)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2021.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 4195
        table[tech.CCGT] = 1643
        table[tech.CCGT_CCS] = 4270
        table[tech.CentralReceiver] = 5530
        table[tech.Coal_CCS] = 8906
        table[tech.OCGT] = 822
        table[tech.PV1Axis] = 624
        table[tech.Wind] = 1830


class GenCost2022:
    """GenCost 2021-22 costs.

    Source:
    CSIRO GenCost 2021-22 report
    https://data.csiro.au/collections/collection/CIcsiro:44228
    """

    lifetime = 30

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        self.discount_rate = discount
        self.ccs_storage_per_t = ccs_price
        # bioenergy costs taken from CSIRO energy storage report for AEMO
        self.bioenergy_price_per_gj = 12
        self.coal_price_per_gj = coal_price
        self.gas_price_per_gj = gas_price
        self.diesel_price_per_litre = 1.50
        self.annuityf = annuity_factor(self.lifetime, discount)

        # Fixed O&M (FOM) costs
        # Note: These are the same for all years (2030, 2040, 2050),
        # so we can set them once here.
        self.fixed_om_costs = {
            tech.Black_Coal: 53.2,
            tech.CCGT: 10.9,
            tech.CCGT_CCS: 16.4,
            tech.CentralReceiver: 120.0,
            tech.Coal_CCS: 77.8,
            tech.Hydro: 0,
            tech.OCGT: 10.2,
            tech.PV1Axis: 17.0,
            tech.PumpedHydro: 0,
            tech.Wind: 25.0,
            tech.WindOffshore: 149.9,
            tech.DemandResponse: 0
        }

        # Variable O&M (VOM) costs
        # Likewise, these are the same for all years (2030, 2040, 2050).
        self.opcost_per_mwh = {
            tech.Black_Coal: 4.2,
            tech.CCGT: 3.7,
            tech.CCGT_CCS: 7.2,
            tech.CentralReceiver: 0,
            tech.Coal_CCS: 8.0,
            tech.Hydro: 0,
            tech.OCGT: 7.3,
            tech.PV1Axis: 0,
            tech.PumpedHydro: 0,
            tech.Wind: 0,
            tech.WindOffshore: 0
        }

        # Common capital costs
        self.capcost_per_kw = {
            tech.Hydro: 0,
            tech.PumpedHydro: 0,
            tech.DemandResponse: 0
        }

        # Storage is expressed on a total cost basis (GenCost 2022, p. 18)
        # Figures are entered in the classes in $/kWh, but these are
        # converted to $/kW in capcost().
        self.totcost_per_kwh = {}


class GenCost2022_2021(GenCost2022):
    """GenCost 2020-21 costs for 2021 (low assumption)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2022.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 4343
        table[tech.CCGT] = 1559
        table[tech.CCGT_CCS] = 4011
        table[tech.CentralReceiver] = 6693
        table[tech.Coal_CCS] = 9077
        table[tech.OCGT] = 873
        table[tech.PV1Axis] = 1441
        table[tech.Wind] = 1960
        table[tech.WindOffshore] = 4649

        table = self.totcost_per_kwh
        table[tech.Battery] = {1: 790, 2: 527, 4: 407, 8: 357}


class GenCost2022_2030_CP(GenCost2022):
    """GenCost 2021-22 costs for 2030 (current policies)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2022.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 4223
        table[tech.CCGT] = 1516
        table[tech.CCGT_CCS] = 3957
        table[tech.CentralReceiver] = 5660
        table[tech.Coal_CCS] = 8884
        table[tech.OCGT] = 741
        table[tech.PV1Axis] = 1013
        table[tech.Wind] = 1897
        table[tech.WindOffshore] = 4545

        table = self.totcost_per_kwh
        table[tech.Battery] = {1: 687, 2: 452, 4: 343, 8: 298}


class GenCost2022_2040_CP(GenCost2022):
    """GenCost 2021-22 costs for 2040 (current policies)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2022.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 4078
        table[tech.CCGT] = 1464
        table[tech.CCGT_CCS] = 3892
        table[tech.CentralReceiver] = 4894
        table[tech.Coal_CCS] = 8650
        table[tech.OCGT] = 716
        table[tech.PV1Axis] = 733
        table[tech.Wind] = 1868
        table[tech.WindOffshore] = 4482

        table = self.totcost_per_kwh
        table[tech.Battery] = {1: 565, 2: 363, 4: 269, 8: 230}


class GenCost2022_2050_CP(GenCost2022):
    """GenCost 2021-22 costs for 2050 (current policies)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2022.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 3937
        table[tech.CCGT] = 1413
        table[tech.CCGT_CCS] = 3324
        table[tech.CentralReceiver] = 4103
        table[tech.Coal_CCS] = 7958
        table[tech.OCGT] = 691
        table[tech.PV1Axis] = 644
        table[tech.Wind] = 1828
        table[tech.WindOffshore] = 4431

        table = self.totcost_per_kwh
        table[tech.Battery] = {1: 485, 2: 315, 4: 236, 8: 203}


class GenCost2022_2030_NZE2050(GenCost2022):
    """GenCost 2021-22 costs for 2030 (Global NZE by 2050)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2022.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 4208
        table[tech.CCGT] = 1511
        table[tech.CCGT_CCS] = 3725
        table[tech.CentralReceiver] = 4657
        table[tech.Coal_CCS] = 8631
        table[tech.OCGT] = 741
        table[tech.PV1Axis] = 785
        table[tech.Wind] = 1633
        table[tech.WindOffshore] = 2967

        table = self.totcost_per_kwh
        table[tech.Battery] = {1: 553, 2: 344, 4: 242, 8: 200}


class GenCost2022_2040_NZE2050(GenCost2022):
    """GenCost 2021-22 costs for 2040 (Global NZE by 2050)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2022.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 4063
        table[tech.CCGT] = 1459
        table[tech.CCGT_CCS] = 3039
        table[tech.CentralReceiver] = 3620
        table[tech.Coal_CCS] = 7768
        table[tech.OCGT] = 716
        table[tech.PV1Axis] = 578
        table[tech.Wind] = 1553
        table[tech.WindOffshore] = 2653

        table = self.totcost_per_kwh
        table[tech.Battery] = {1: 436, 2: 272, 4: 194, 8: 161}


class GenCost2022_2050_NZE2050(GenCost2022):
    """GenCost 2021-22 costs for 2050 (Global NZE by 2050)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2022.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 3930
        table[tech.CCGT] = 1411
        table[tech.CCGT_CCS] = 2978
        table[tech.CentralReceiver] = 2911
        table[tech.Coal_CCS] = 7552
        table[tech.OCGT] = 691
        table[tech.PV1Axis] = 521
        table[tech.Wind] = 1521
        table[tech.WindOffshore] = 2506

        table = self.totcost_per_kwh
        table[tech.Battery] = {1: 337, 2: 220, 4: 167, 8: 144}


class GenCost2022_2030_NZEPost2050(GenCost2022):
    """GenCost 2021-22 costs for 2030 (Global NZE post 2050)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2022.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 4223
        table[tech.CCGT] = 1516
        table[tech.CCGT_CCS] = 3784
        table[tech.CentralReceiver] = 5236
        table[tech.Coal_CCS] = 8747
        table[tech.OCGT] = 741
        table[tech.PV1Axis] = 1046
        table[tech.Wind] = 1778
        table[tech.WindOffshore] = 4437

        table = self.totcost_per_kwh
        table[tech.Battery] = {1: 608, 2: 390, 4: 287, 8: 244}


class GenCost2022_2040_NZEPost2050(GenCost2022):
    """GenCost 2021-22 costs for 2040 (Global NZE post 2050)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2022.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 4078
        table[tech.CCGT] = 1464
        table[tech.CCGT_CCS] = 3276
        table[tech.CentralReceiver] = 4181
        table[tech.Coal_CCS] = 8025
        table[tech.OCGT] = 716
        table[tech.PV1Axis] = 689
        table[tech.Wind] = 1648
        table[tech.WindOffshore] = 3772

        table = self.totcost_per_kwh
        table[tech.Battery] = {1: 483, 2: 309, 4: 227, 8: 193}


class GenCost2022_2050_NZEPost2050(GenCost2022):
    """GenCost 2021-22 costs for 2050 (Global NZE post 2050)."""

    def __init__(self, discount, coal_price, gas_price, ccs_price):
        """Construct a cost object."""
        GenCost2022.__init__(self, discount, coal_price, gas_price, ccs_price)
        table = self.capcost_per_kw
        table[tech.Black_Coal] = 3937
        table[tech.CCGT] = 1413
        table[tech.CCGT_CCS] = 3206
        table[tech.CentralReceiver] = 3478
        table[tech.Coal_CCS] = 7792
        table[tech.OCGT] = 691
        table[tech.PV1Axis] = 530
        table[tech.Wind] = 1546
        table[tech.WindOffshore] = 3168

        table = self.totcost_per_kwh
        table[tech.Battery] = {1: 385, 2: 255, 4: 196, 8: 172}


cost_scenarios = {'Null': NullCosts,
                  'AETA2012-in2030-low': AETA2012_2030Low,
                  'AETA2012-in2030-mid': AETA2012_2030Mid,
                  'AETA2012-in2030-high': AETA2012_2030High,
                  'AETA2013-in2030-low': AETA2013_2030Low,
                  'AETA2013-in2030-mid': AETA2013_2030Mid,
                  'AETA2013-in2030-high': AETA2013_2030High,
                  'CEEM2016-in2030': CEEM2016_2030,
                  'GenCost2021-in2020': GenCost2021_2020,
                  'GenCost2021-in2030-low': GenCost2021_2030Low,
                  'GenCost2021-in2030-high': GenCost2021_2030High,
                  'GenCost2021-in2040-low': GenCost2021_2040Low,
                  'GenCost2021-in2040-high': GenCost2021_2040High,
                  'GenCost2021-in2050-low': GenCost2021_2050Low,
                  'GenCost2021-in2050-high': GenCost2021_2050High,
                  'GenCost2022-in2021': GenCost2022_2021,
                  'GenCost2022-in2030-CP': GenCost2022_2030_CP,
                  'GenCost2022-in2030-NZE2050': GenCost2022_2030_NZE2050,
                  'GenCost2022-in2030-NZE2050+': GenCost2022_2030_NZEPost2050,
                  'GenCost2022-in2040-CP': GenCost2022_2040_CP,
                  'GenCost2022-in2040-NZE2050': GenCost2022_2040_NZE2050,
                  'GenCost2022-in2040-NZE2050+': GenCost2022_2040_NZEPost2050,
                  'GenCost2022-in2050-CP': GenCost2022_2050_CP,
                  'GenCost2022-in2050-NZE2050': GenCost2022_2050_NZE2050,
                  'GenCost2022-in2050-NZE2050+': GenCost2022_2050_NZEPost2050,
                  'PGTR2015': APGTR2015,
                  'PGTR2030': APGTR2030}



# Copyright (C) 2011, 2014 Ben Elliston
# Copyright (C) 2014, 2015 The University of New South Wales
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""Market regions consisting of one or more polygons."""


class Region():
    """Each region is described by a Region object."""

    def __init__(self, ordinal, regionid, descr):
        """
        Construct a Region given an ordinal number, region ID and description.

        >>> r = Region(0, 'cbr', 'Capital region')
        """
        self.id = regionid
        self.descr = descr
        self.num = ordinal
        self.polygons = None

    def __repr__(self):
        """Return region code."""
        return self.id

    def __index__(self):
        """Return region number."""
        return self.num

    def __copy__(self):
        """Prevent copying."""
        return self

    def __deepcopy__(self, _):
        """Prevent deepcopying."""
        return self


nsw = Region(0, 'NSW1', 'New South Wales')
qld = Region(1, 'QLD1', 'Queensland')
sa = Region(2, 'SA1', 'South Australia')
snowy = Region(3, 'SNOWY1', 'Snowy Mountains')
tas = Region(4, 'TAS1', 'Tasmania')
vic = Region(5, 'VIC1', 'Victoria')
All = [nsw, qld, sa, snowy, tas, vic]
NUMREGIONS = len(All)



# Copyright (C) 2022 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# pylint: disable=protected-access

"""A testsuite for the costs module."""

import unittest

from nemo import costs


class TestCosts(unittest.TestCase):
    """Test costs.py."""

    def setUp(self):
        """Test harness setup."""
        self.discount = 0.05
        self.coal_price = 2.00
        self.gas_price = 9.00
        self.ccs_price = 27
        self.costclasses = costs.cost_scenarios.items()

    def test_annuity_factor(self):
        """Test annuity_factor function."""
        result = round(costs.annuity_factor(30, 0.05), 3)
        self.assertEqual(result, 15.372)

    def test_zero_dict(self):
        """Test _ZeroDict dictionary type."""
        zerodict = costs.NullCosts._ZeroDict()
        self.assertEqual(zerodict[123], 0)

    def test_table(self):
        """Check all table entries are valid."""
        for _, cls in self.costclasses:
            obj = cls(self.discount, self.coal_price, self.gas_price,
                      self.ccs_price)
            if isinstance(obj, costs.NullCosts):
                # special case for NullCosts
                self.assertEqual(obj.coal_price_per_gj, 0)
                self.assertEqual(obj.gas_price_per_gj, 0)
                self.assertEqual(obj.ccs_storage_per_t, 0)
            else:
                self.assertEqual(obj.coal_price_per_gj, self.coal_price)
                self.assertEqual(obj.gas_price_per_gj, self.gas_price)
                self.assertEqual(obj.ccs_storage_per_t, self.ccs_price)

    def test_costs_sensible(self):
        """Test if cost values are sensible."""
        for _, cls in self.costclasses:
            obj = cls(self.discount, self.coal_price, self.gas_price,
                      self.ccs_price)

            for table in [obj.capcost_per_kw, obj.fixed_om_costs,
                          obj.opcost_per_mwh]:
                self.assertTrue(all(value >= 0 for value in table.values()))



"""Setup script."""

import time

from setuptools import setup

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(name='nemopt',
      # Use a datestamp as the version.
      version=time.strftime('%Y%m%d'),
      packages=['awklite', 'nemo'],
      description='National Electricity Market Optimiser',
      author='Ben Elliston',
      author_email='bje@air.net.au',
      license='GPLv3',
      url='https://nemo.ozlabs.org',
      long_description=long_description,
      long_description_content_type='text/markdown',
      keywords=['electricity', 'model', 'scenarios'],
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Environment :: X11 Applications',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Programming Language :: Python :: 3 :: Only',
          'Topic :: Scientific/Engineering',
      ],
      data_files=[('etc', ['nemo.cfg'])],
      scripts=['evolve', 'replay', 'summary'],
      install_requires=[
          'deap',
          'Gooey>=1.0.4',
          'matplotlib',
          'numpy',
          'pandas',
          'pint',
          'requests'])
